import logging
import os
import warnings
from collections import UserDict
from contextlib import contextmanager
from typing import Any, Dict, List, MutableSequence, Optional, Sequence, Tuple, Union

import torch
import tqdm
from packaging import version
from torch import Tensor
from torch.utils.data import DataLoader
from transformers.utils import ModelOutput

from pytorch_ie.core.document import Document
from pytorch_ie.core.model import PyTorchIEModel
from pytorch_ie.core.taskmodule import (
    InplaceNotSupportedException,
    TaskEncoding,
    TaskEncodingDataset,
    TaskModule,
    TaskOutput,
)

logger = logging.getLogger(__name__)


class Pipeline:
    """
    The Pipeline class is the class from which all pipelines inherit. Refer to this class for methods shared across
    different pipelines.

    Base class implementing pipelined operations. Pipeline workflow is defined as a sequence of the following
    operations:

        Input -> Tokenization -> Model Inference -> Post-Processing (task dependent) -> Output

    Pipeline supports running on CPU or GPU through the device argument (see below).

    Some pipeline, like for instance :class:`~transformers.FeatureExtractionPipeline` (:obj:`'feature-extraction'` )
    output large tensor object as nested-lists. In order to avoid dumping such large structure as textual data we
    provide the :obj:`binary_output` constructor argument. If set to :obj:`True`, the output will be stored in the
    pickle format.
    """

    default_input_names = None

    def __init__(
        self,
        model: PyTorchIEModel,
        taskmodule: TaskModule,
        # args_parser: ArgumentHandler = None,
        device: int = -1,
        binary_output: bool = False,
        **kwargs,
    ):
        self.taskmodule = taskmodule
        self.device = torch.device("cpu" if device < 0 else f"cuda:{device}")
        self.binary_output = binary_output

        # Module.to() returns just self, but moved to the device. This is not correctly
        # reflected in typing of PyTorch.
        self.model: PyTorchIEModel = model.to(self.device)  # type: ignore

        self.call_count = 0
        (
            self._preprocess_params,
            self._dataloader_params,
            self._forward_params,
            self._postprocess_params,
        ) = self._sanitize_parameters(**kwargs)

    def save_pretrained(self, save_directory: str):
        """
        Save the pipeline's model and taskmodule.

        Args:
            save_directory (:obj:`str`):
                A path to the directory where to saved. It will be created if it doesn't exist.
        """
        if os.path.isfile(save_directory):
            logger.error(f"Provided path ({save_directory}) should be a directory, not a file")
            return
        os.makedirs(save_directory, exist_ok=True)

        self.model.save_pretrained(save_directory)

        self.taskmodule.save_pretrained(save_directory)

    def transform(self, X):
        """
        Scikit / Keras interface to transformers' pipelines. This method will forward to __call__().
        """
        return self(X=X)

    def predict(self, X):
        """
        Scikit / Keras interface to transformers' pipelines. This method will forward to __call__().
        """
        return self(X=X)

    @contextmanager
    def device_placement(self):
        """
        Context Manager allowing tensor allocation on the user-specified device in framework agnostic way.

        Returns:
            Context manager

        Examples::

            # Explicitly ask for tensor allocation on CUDA device :0
            pipe = pipeline(..., device=0)
            with pipe.device_placement():
                # Every framework specific tensor allocation will be done on the request device
                output = pipe(...)
        """
        if self.device.type == "cuda":
            torch.cuda.set_device(self.device)

        yield

    def ensure_tensor_on_device(self, **inputs):
        """
        Ensure PyTorch tensors are on the specified device.

        Args:
            inputs (keyword arguments that should be :obj:`torch.Tensor`, the rest is ignored): The tensors to place on :obj:`self.device`.
            Recursive on lists **only**.

        Return:
            :obj:`Dict[str, torch.Tensor]`: The same as :obj:`inputs` but on the proper device.
        """
        return self._ensure_tensor_on_device(inputs, self.device)

    def _ensure_tensor_on_device(self, inputs, device):
        if isinstance(inputs, ModelOutput):
            return type(inputs)(
                {
                    name: self._ensure_tensor_on_device(tensor, device)
                    for name, tensor in inputs.items()
                }
            )
        elif isinstance(inputs, dict):
            return {
                name: self._ensure_tensor_on_device(tensor, device)
                for name, tensor in inputs.items()
            }
        elif isinstance(inputs, UserDict):
            return UserDict(
                {
                    name: self._ensure_tensor_on_device(tensor, device)
                    for name, tensor in inputs.items()
                }
            )
        elif isinstance(inputs, list):
            return [self._ensure_tensor_on_device(item, device) for item in inputs]
        elif isinstance(inputs, tuple):
            return tuple(self._ensure_tensor_on_device(item, device) for item in inputs)
        elif isinstance(inputs, torch.Tensor):
            return inputs.to(device)
        else:
            return inputs

    def _sanitize_parameters(
        self, **pipeline_parameters
    ) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
        """
        _sanitize_parameters will be called with any excessive named arguments from either `__init__` or `__call__`
        methods. It should return 4 dictionaries of the resolved parameters used by the various `preprocess`,
        `get_dataloader`, `forward` and `postprocess` methods. Do not fill dictionaries if the caller didn't specify
        a kwargs. This lets you keep defaults in function signatures, which is more "natural".

        It is not meant to be called directly, it will be automatically called and the final parameters resolved by
        `__init__` and `__call__`
        """
        preprocess_parameters = {}
        dataloader_params = {}
        forward_parameters = {}
        postprocess_parameters: Dict[str, Any] = {}

        # set preprocess parameters
        for p_name in ["document_batch_size"]:
            if p_name in pipeline_parameters:
                preprocess_parameters[p_name] = pipeline_parameters[p_name]

        # set forward parameters
        for p_name in ["show_progress_bar", "fast_dev_run"]:
            if p_name in pipeline_parameters:
                forward_parameters[p_name] = pipeline_parameters[p_name]

        # set dataloader parameters
        for p_name in ["batch_size", "num_workers", "shuffle"]:
            if p_name in pipeline_parameters:
                dataloader_params[p_name] = pipeline_parameters[p_name]

        # set postprocess parameters
        for p_name in ["inplace"]:
            if p_name in pipeline_parameters:
                postprocess_parameters[p_name] = pipeline_parameters[p_name]

        return preprocess_parameters, dataloader_params, forward_parameters, postprocess_parameters

    def preprocess(
        self,
        documents: Sequence[Document],
        document_batch_size: Optional[int] = None,
        **preprocess_parameters: Dict,
    ) -> Sequence[TaskEncoding]:
        """
        Preprocess will take the `input_` of a specific pipeline and return a dictionnary of everything necessary for
        `_forward` to run properly. It should contain at least one tensor, but might have arbitrary other items.
        """

        encodings = self.taskmodule.encode(
            documents,
            encode_target=False,
            as_task_encoding_sequence=True,
            document_batch_size=document_batch_size,
        )
        if not isinstance(encodings, Sequence):
            raise TypeError("preprocess has to return a sequence")
        return encodings

    def _forward(
        self, input_tensors: Tuple[Dict[str, Tensor], Any, Any, Any], **forward_parameters: Dict
    ) -> Dict:
        """
        _forward will receive the prepared dictionnary from `preprocess` and run it on the model. This method might
        involve the GPU or the CPU and should be agnostic to it. Isolating this function is the reason for `preprocess`
        and `postprocess` to exist, so that the hot path, this method generally can run as fast as possible.

        It is not meant to be called directly, `forward` is preferred. It is basically the same but contains additional
        code surrounding `_forward` making sure tensors and models are on the same device, disabling the training part
        of the code (leading to faster inference).
        """
        inputs = input_tensors[0]
        return self.model.predict(inputs, **forward_parameters)

    def postprocess(
        self,
        model_inputs: Sequence[TaskEncoding],
        model_outputs: Sequence[TaskOutput],
        **postprocess_parameters,
    ) -> Sequence[Document]:
        """
        Postprocess will receive the model inputs and (unbatched) model outputs and reformat them into
        something more friendly. Generally it will output a list of documents.
        """
        # This creates annotations from the model outputs and attaches them to the correct documents.
        return self.taskmodule.decode(
            task_encodings=model_inputs,
            task_outputs=model_outputs,
            **postprocess_parameters,
        )

    def get_inference_context(self):
        inference_context = (
            torch.inference_mode
            if version.parse(torch.__version__) >= version.parse("1.9.0")
            else torch.no_grad
        )
        return inference_context

    def forward(self, model_inputs, **forward_params):
        with self.device_placement():
            inference_context = self.get_inference_context()
            with inference_context():
                model_inputs = self._ensure_tensor_on_device(model_inputs, device=self.device)
                model_outputs = self._forward(model_inputs, **forward_params)
                model_outputs = self._ensure_tensor_on_device(
                    model_outputs, device=torch.device("cpu")
                )
        return model_outputs

    def get_dataloader(
        self,
        model_inputs: Sequence[TaskEncoding],
        batch_size: int = 1,
        num_workers: int = 8,
        **kwargs,
    ):
        dataloader: DataLoader[TaskEncoding] = DataLoader(
            TaskEncodingDataset(model_inputs),
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            collate_fn=self.taskmodule.collate,
            **kwargs,
        )

        return dataloader

    def __call__(
        self,
        documents: Union[Document, Sequence[Document]],
        *args,
        **kwargs,
    ) -> Union[Document, Sequence[Document]]:
        if args:
            logger.warning(f"Ignoring args : {args}")
        (
            preprocess_params,
            dataloader_params,
            forward_params,
            postprocess_params,
        ) = self._sanitize_parameters(**kwargs)

        in_place: bool = postprocess_params.get("inplace", True)
        if in_place and not isinstance(documents, (MutableSequence, Document)):
            raise InplaceNotSupportedException(
                "Immutable sequences of Documents (such as Datasets) can't be modified in place. Please set inplace=False."
            )

        if "TOKENIZERS_PARALLELISM" not in os.environ:
            logger.info(
                "Disabling tokenizer parallelism, we're using DataLoader multithreading already"
            )
            os.environ["TOKENIZERS_PARALLELISM"] = "false"

        # Fuse __init__ params and __call__ params without modifying the __init__ ones.
        preprocess_params = {**self._preprocess_params, **preprocess_params}
        dataloader_params = {**self._dataloader_params, **dataloader_params}
        forward_params = {**self._forward_params, **forward_params}
        postprocess_params = {**self._postprocess_params, **postprocess_params}

        self.call_count += 1
        if self.call_count > 10 and self.device.type == "cuda":
            warnings.warn(
                "You seem to be using the pipelines sequentially on GPU. In order to maximize efficiency please use a dataset",
                UserWarning,
            )

        single_document = False
        if isinstance(documents, Document):
            single_document = True
            documents = [documents]

        # This creates encodings from the documents. It modifies the documents and may produce multiple entries per
        # document.
        model_inputs = self.preprocess(documents, **preprocess_params)
        if forward_params.pop("fast_dev_run", False):
            warnings.warn(
                "Execute a fast dev run, only the first two model inputs will be processed."
            )
            model_inputs = model_inputs[:2]
        # Create a dataloader from the model inputs. This uses taskmodule.collate().
        dataloader = self.get_dataloader(model_inputs=model_inputs, **dataloader_params)

        show_progress_bar = forward_params.pop("show_progress_bar", False)
        model_outputs: List = []
        with torch.no_grad():
            for batch in tqdm.tqdm(dataloader, desc="inference", disable=not show_progress_bar):
                output = self.forward(batch, **forward_params)
                processed_output = self.taskmodule.unbatch_output(output)
                model_outputs.extend(processed_output)

        assert len(model_inputs) == len(
            model_outputs
        ), f"length mismatch: len(model_inputs) [{len(model_inputs)}] != len(model_outputs) [{len(model_outputs)}]"

        documents = self.postprocess(
            model_inputs=model_inputs,
            model_outputs=model_outputs,
            **postprocess_params,
        )
        if single_document:
            return documents[0]
        else:
            return documents



import inspect
from collections import defaultdict
from typing import Dict, Optional, Type, TypeVar


class RegistrationError(Exception):
    pass


T = TypeVar("T", bound="Registrable")


class Registrable:
    _registry: Dict[Type, Dict[str, Type]] = defaultdict(dict)

    @classmethod
    def register(
        cls: Type[T],
        name: Optional[str] = None,
    ):
        registry = Registrable._registry[cls]

        def add_subclass_to_registry(subclass: Type[T]) -> Type[T]:
            register_name = subclass.__name__ if name is None else name

            if not inspect.isclass(subclass) or not issubclass(subclass, cls):
                raise RegistrationError(
                    f"Cannot register {subclass.__name__} as {register_name}; "
                    f"{subclass.__name__} must be a subclass of {cls.__name__}"
                )

            if register_name in registry:
                raise RegistrationError(
                    f"Cannot register {subclass.__name__} as {register_name}; "
                    f"name already in use for {registry[register_name].__name__}"
                )

            registry[register_name] = subclass
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls: Type[T], name: str) -> Type[T]:
        if name in Registrable._registry[cls]:
            return Registrable._registry[cls][name]

        raise RegistrationError(f"{name} is not a registered name for {cls.__name__}.")

    @classmethod
    def registered_name_for_class(cls: Type[T], clazz: Type[T]) -> Optional[str]:
        inverse_lookup = {v: k for k, v in Registrable._registry[cls].items()}
        return inverse_lookup.get(clazz)

    @classmethod
    def name_for_object_class(cls: Type[T], obj: T) -> str:
        obj_class = obj.__class__
        registered_name = cls.registered_name_for_class(obj_class)
        return registered_name if registered_name is not None else obj_class.__name__



"""
workflow:
    Document
        -> (InputEncoding, TargetEncoding) -> TaskEncoding -> TaskBatchEncoding
            -> ModelBatchEncoding -> ModelBatchOutput
        -> TaskOutput
    -> Document
"""

import logging
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypedDict,
    Union,
)

import numpy as np
import torch
from transformers import AutoTokenizer
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import TruncationStrategy
from typing_extensions import TypeAlias

from pytorch_ie.annotations import (
    BinaryRelation,
    LabeledSpan,
    MultiLabeledBinaryRelation,
    NaryRelation,
    Span,
)
from pytorch_ie.core import AnnotationLayer, Document, TaskEncoding, TaskModule
from pytorch_ie.documents import (
    TextDocument,
    TextDocumentWithLabeledSpansAndBinaryRelations,
    TextDocumentWithLabeledSpansBinaryRelationsAndLabeledPartitions,
)
from pytorch_ie.models.transformer_text_classification import ModelOutputType, ModelStepInputType
from pytorch_ie.taskmodules.interface import ChangesTokenizerVocabSize
from pytorch_ie.utils.span import get_token_slice, is_contained_in
from pytorch_ie.utils.window import get_window_around_slice

InputEncodingType: TypeAlias = Dict[str, Any]
TargetEncodingType: TypeAlias = Sequence[int]

TaskEncodingType: TypeAlias = TaskEncoding[
    TextDocument,
    InputEncodingType,
    TargetEncodingType,
]


class TaskOutputType(TypedDict, total=False):
    labels: Sequence[str]
    probabilities: Sequence[float]


TaskModuleType: TypeAlias = TaskModule[
    TextDocument,
    InputEncodingType,
    TargetEncodingType,
    ModelStepInputType,
    ModelOutputType,
    TaskOutputType,
]


HEAD = "head"
TAIL = "tail"
START = "start"
END = "end"


logger = logging.getLogger(__name__)


class RelationArgument:
    def __init__(
        self,
        entity: LabeledSpan,
        role: str,
        token_span: Span,
        add_type_to_marker: bool,
        role_to_marker: Dict[str, str],
    ) -> None:
        self.entity = entity
        self.role_to_marker = role_to_marker
        if role not in self.role_to_marker:
            raise Exception(f"role={role} not in role_to_marker={role_to_marker}")
        self.role = role
        self.token_span = token_span
        self.add_type_to_marker = add_type_to_marker

    @property
    def as_start_marker(self) -> str:
        return self._get_marker(is_start=True)

    @property
    def as_end_marker(self) -> str:
        return self._get_marker(is_start=False)

    @property
    def role_marker(self) -> str:
        return self.role_to_marker[self.role]

    def _get_marker(self, is_start: bool = True) -> str:
        return f"[{'' if is_start else '/'}{self.role_marker}" + (
            f":{self.entity.label}]" if self.add_type_to_marker else "]"
        )

    @property
    def as_append_marker(self) -> str:
        return f"[{self.role_marker}={self.entity.label}]"

    def shift_token_span(self, value: int):
        self.token_span = Span(
            start=self.token_span.start + value, end=self.token_span.end + value
        )


@TaskModule.register()
class TransformerRETextClassificationTaskModule(TaskModuleType, ChangesTokenizerVocabSize):
    """Marker based relation extraction. This taskmodule prepares the input token ids in such a way
    that before and after the candidate head and tail entities special marker tokens are inserted.
    Then, the modified token ids can be simply passed into a transformer based text classifier
    model.

    parameters:

        partition_annotation: str, optional. If specified, LabeledSpan annotations with this name are
            expected to define partitions of the document that will be processed individually, e.g. sentences
            or sections of the document text.
        none_label: str, defaults to "no_relation". The relation label that indicate dummy/negative relations.
            Predicted relations with that label will not be added to the document(s).
        max_window: int, optional. If specified, use the tokens in a window of maximal this amount of tokens
            around the center of head and tail entities and pass only that into the transformer.
        create_relation_candidates: bool, defaults to False. If True, create relation candidates by pairwise
            combining all entities in the document and assigning the none_label. If the document already contains
            a relation with the entity pair, we do not add it again. If False, assume that the document already
            contains relation annotations including negative examples (i.e. relations with the none_label).
    """

    PREPARED_ATTRIBUTES = ["label_to_id", "entity_labels"]

    def __init__(
        self,
        tokenizer_name_or_path: str,
        # this is deprecated, the target of the relation layer already specifies the entity layer
        entity_annotation: Optional[str] = None,
        relation_annotation: str = "binary_relations",
        create_relation_candidates: bool = False,
        partition_annotation: Optional[str] = None,
        none_label: str = "no_relation",
        padding: Union[bool, str, PaddingStrategy] = True,
        truncation: Union[bool, str, TruncationStrategy] = True,
        max_length: Optional[int] = None,
        pad_to_multiple_of: Optional[int] = None,
        multi_label: bool = False,
        label_to_id: Optional[Dict[str, int]] = None,
        add_type_to_marker: bool = False,
        argument_role_to_marker: Optional[Dict[str, str]] = None,
        single_argument_pair: bool = True,
        append_markers: bool = False,
        entity_labels: Optional[List[str]] = None,
        reversed_relation_label_suffix: Optional[str] = None,
        max_window: Optional[int] = None,
        log_first_n_examples: int = 0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if entity_annotation is not None:
            logger.warning(
                "The parameter entity_annotation is deprecated and will be discarded because it is not necessary "
                "anymore. The target of the relation layer already specifies the entity layer."
            )
        self.save_hyperparameters(ignore=["entity_annotation"])

        self.relation_annotation = relation_annotation
        self.create_relation_candidates = create_relation_candidates
        self.padding = padding
        self.truncation = truncation
        self.label_to_id = label_to_id or {}
        self.id_to_label = {v: k for k, v in self.label_to_id.items()}
        self.max_length = max_length
        self.pad_to_multiple_of = pad_to_multiple_of
        self.multi_label = multi_label
        self.add_type_to_marker = add_type_to_marker
        self.single_argument_pair = single_argument_pair
        self.append_markers = append_markers
        self.entity_labels = entity_labels
        self.partition_annotation = partition_annotation
        self.none_label = none_label
        self.reversed_relation_label_suffix = reversed_relation_label_suffix
        self.max_window = max_window
        # overwrite None with 0 for backward compatibility
        self.log_first_n_examples = log_first_n_examples or 0

        if argument_role_to_marker is None:
            self.argument_role_to_marker = {HEAD: "H", TAIL: "T"}
        else:
            self.argument_role_to_marker = argument_role_to_marker

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

        self.argument_markers = None

        self._logged_examples_counter = 0

    @property
    def document_type(self) -> Optional[Type[TextDocument]]:
        dt: Type[TextDocument]
        if self.partition_annotation is not None:
            dt = TextDocumentWithLabeledSpansBinaryRelationsAndLabeledPartitions
        else:
            dt = TextDocumentWithLabeledSpansAndBinaryRelations

        if self.relation_annotation == "binary_relations":
            return dt
        else:
            logger.warning(
                f"relation_annotation={self.relation_annotation} is "
                f"not the default value ('binary_relations'), so the taskmodule {type(self).__name__} can not request "
                f"the usual document type for auto-conversion ({dt.__name__}) because this has the bespoken default "
                f"value as layer name instead of the provided one."
            )
            return None

    def get_relation_layer(self, document: Document) -> AnnotationLayer[BinaryRelation]:
        return document[self.relation_annotation]

    def get_entity_layer(self, document: Document) -> AnnotationLayer[LabeledSpan]:
        relations: AnnotationLayer[BinaryRelation] = self.get_relation_layer(document)
        if len(relations._targets) != 1:
            raise Exception(
                f"the relation layer is expected to target exactly one entity layer, but it has "
                f"the following targets: {relations._targets}"
            )
        entity_layer_name = relations._targets[0]
        return document[entity_layer_name]

    def _prepare(self, documents: Sequence[TextDocument]) -> None:
        entity_labels: Set[str] = set()
        relation_labels: Set[str] = set()
        for document in documents:
            relations: AnnotationLayer[BinaryRelation] = self.get_relation_layer(document)
            entities: AnnotationLayer[LabeledSpan] = self.get_entity_layer(document)

            for entity in entities:
                entity_labels.add(entity.label)

            for relation in relations:
                relation_labels.add(relation.label)

        if self.none_label in relation_labels:
            relation_labels.remove(self.none_label)

        self.label_to_id = {label: i + 1 for i, label in enumerate(sorted(relation_labels))}
        self.label_to_id[self.none_label] = 0

        self.entity_labels = sorted(entity_labels)

    def construct_argument_markers(self) -> List[str]:
        # ignore the typing because we know that this is only called on a prepared taskmodule,
        # i.e. self.entity_labels is already set by _prepare or __init__
        entity_labels: List[str] = self.entity_labels  # type: ignore
        argument_markers: Set[str] = set()
        for arg_role, role_marker in self.argument_role_to_marker.items():
            for arg_pos in [START, END]:
                is_start = arg_pos == START
                argument_markers.add(f"[{'' if is_start else '/'}{role_marker}]")
                if self.add_type_to_marker:
                    for entity_type in entity_labels:
                        argument_markers.add(
                            f"[{'' if is_start else '/'}{role_marker}"
                            f"{':' + entity_type if self.add_type_to_marker else ''}]"
                        )
                if self.append_markers:
                    for entity_type in entity_labels:
                        argument_markers.add(f"[{role_marker}={entity_type}]")

        return sorted(list(argument_markers))

    def _post_prepare(self):
        self.argument_markers = self.construct_argument_markers()
        self.tokenizer.add_tokens(self.argument_markers, special_tokens=True)

        self.argument_markers_to_id = {
            marker: self.tokenizer.vocab[marker] for marker in self.argument_markers
        }
        self.sep_token_id = self.tokenizer.vocab[self.tokenizer.sep_token]

        self.id_to_label = {v: k for k, v in self.label_to_id.items()}

    def _create_relation_candidates(
        self,
        document: Document,
    ) -> List[BinaryRelation]:
        relation_candidates: List[BinaryRelation] = []
        relations: AnnotationLayer[BinaryRelation] = self.get_relation_layer(document)
        entities: AnnotationLayer[LabeledSpan] = self.get_entity_layer(document)
        arguments_to_relation = {(rel.head, rel.tail): rel for rel in relations}
        # iterate over all possible argument candidates
        for head in entities:
            for tail in entities:
                if head != tail:
                    # If there is no relation with the candidate arguments, we create a relation candidate with the
                    # none label. Otherwise, we use the existing relation.
                    candidate = arguments_to_relation.get(
                        (head, tail),
                        BinaryRelation(
                            head=head,
                            tail=tail,
                            label=self.none_label,
                            score=1.0,
                        ),
                    )
                    relation_candidates.append(candidate)

        return relation_candidates

    def encode_input(
        self,
        document: TextDocument,
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType],]]:
        relations: Sequence[BinaryRelation]
        if self.create_relation_candidates:
            relations = self._create_relation_candidates(document)
        else:
            relations = self.get_relation_layer(document)

        partitions: Sequence[Span]
        if self.partition_annotation is not None:
            partitions = document[self.partition_annotation]
            if len(partitions) == 0:
                logger.warning(
                    f"the document {document.id} has no '{self.partition_annotation}' partition entries, "
                    f"no inputs will be created!"
                )
        else:
            # use single dummy partition
            partitions = [Span(start=0, end=len(document.text))]

        task_encodings: List[TaskEncodingType] = []
        for partition in partitions:
            without_special_tokens = self.max_window is not None
            text = document.text[partition.start : partition.end]
            encoding = self.tokenizer(
                text,
                padding=False,
                truncation=self.truncation if self.max_window is None else False,
                max_length=self.max_length,
                is_split_into_words=False,
                return_offsets_mapping=False,
                add_special_tokens=not without_special_tokens,
            )

            for rel in relations:
                arg_spans: List[LabeledSpan]
                if isinstance(rel, BinaryRelation):
                    if not isinstance(rel.head, LabeledSpan) or not isinstance(
                        rel.tail, LabeledSpan
                    ):
                        raise ValueError(
                            f"the taskmodule expects the relation arguments to be of type LabeledSpan, "
                            f"but got {type(rel.head)} and {type(rel.tail)}"
                        )
                    arg_spans = [rel.head, rel.tail]
                    arg_roles = [HEAD, TAIL]
                elif isinstance(rel, NaryRelation):
                    if any(not isinstance(arg, LabeledSpan) for arg in rel.arguments):
                        raise ValueError(
                            f"the taskmodule expects the relation arguments to be of type LabeledSpan, "
                            f"but got {[type(arg) for arg in rel.arguments]}"
                        )
                    arg_spans = list(rel.arguments)
                    arg_roles = list(rel.roles)
                else:
                    raise NotImplementedError(
                        f"the taskmodule does not yet support relations of type: {type(rel)}"
                    )

                # check if the argument spans are in the current partition
                if any(
                    not is_contained_in((arg.start, arg.end), (partition.start, partition.end))
                    for arg in arg_spans
                ):
                    continue

                # map character spans to token spans
                arg_token_slices_including_none = [
                    get_token_slice(
                        character_slice=(arg.start, arg.end),
                        char_to_token_mapper=encoding.char_to_token,
                        character_offset=partition.start,
                    )
                    for arg in arg_spans
                ]
                # Check if the mapping was successful. It may fail (and is None) if any argument start or end does not
                # match a token start or end, respectively.
                if any(token_slice is None for token_slice in arg_token_slices_including_none):
                    logger.warning(
                        f"Skipping invalid example {document.id}, cannot get argument token slice(s)"
                    )
                    continue

                # ignore the typing, because we checked for None above
                arg_token_slices: List[Tuple[int, int]] = arg_token_slices_including_none  # type: ignore

                # create the argument objects
                args = [
                    RelationArgument(
                        entity=span,
                        role=role,
                        token_span=Span(start=token_slice[0], end=token_slice[1]),
                        add_type_to_marker=self.add_type_to_marker,
                        role_to_marker=self.argument_role_to_marker,
                    )
                    for span, role, token_slice in zip(arg_spans, arg_roles, arg_token_slices)
                ]

                input_ids = encoding["input_ids"]

                # windowing: we restrict the input to a window of a maximal size (max_window) with the arguments
                # of the candidate relation in the center (as much as possible)
                if self.max_window is not None:
                    # The actual number of tokens needs to be lower than max_window because we add two
                    # marker tokens (before / after) each argument and the default special tokens
                    # (e.g. CLS and SEP).
                    max_tokens = (
                        self.max_window
                        - len(args) * 2
                        - self.tokenizer.num_special_tokens_to_add()
                    )
                    # if we add the markers also to the end, this decreases the available window again by
                    # two tokens (marker + sep) per argument
                    if self.append_markers:
                        max_tokens -= len(args) * 2
                    # the slice from the beginning of the first entity to the end of the second is required
                    slice_required = (
                        min(arg.token_span.start for arg in args),
                        max(arg.token_span.end for arg in args),
                    )
                    window_slice = get_window_around_slice(
                        slice=slice_required,
                        max_window_size=max_tokens,
                        available_input_length=len(input_ids),
                    )
                    # this happens if slice_required (all arguments) does not fit into max_tokens (the available window)
                    if window_slice is None:
                        continue

                    window_start, window_end = window_slice
                    input_ids = input_ids[window_start:window_end]

                    for arg in args:
                        arg.shift_token_span(-window_start)

                # collect all markers with their target positions
                marker_ids_with_positions = []
                for arg in args:
                    marker_ids_with_positions.append(
                        (self.argument_markers_to_id[arg.as_start_marker], arg.token_span.start)
                    )
                    marker_ids_with_positions.append(
                        (self.argument_markers_to_id[arg.as_end_marker], arg.token_span.end)
                    )

                # create new input ids with the markers inserted
                input_ids_with_markers = list(input_ids)
                offset = 0
                for marker_id, token_position in sorted(
                    marker_ids_with_positions, key=lambda id_pos: id_pos[1]
                ):
                    input_ids_with_markers = (
                        input_ids_with_markers[: token_position + offset]
                        + [marker_id]
                        + input_ids_with_markers[token_position + offset :]
                    )
                    offset += 1

                if self.append_markers:
                    for arg in args:
                        if without_special_tokens:
                            input_ids_with_markers.append(self.sep_token_id)
                            input_ids_with_markers.append(
                                self.argument_markers_to_id[arg.as_append_marker]
                            )
                        else:
                            input_ids_with_markers.append(
                                self.argument_markers_to_id[arg.as_append_marker]
                            )
                            input_ids_with_markers.append(self.sep_token_id)

                # when windowing is used, we have to add the special tokens manually
                if without_special_tokens:
                    input_ids_with_markers = self.tokenizer.build_inputs_with_special_tokens(
                        token_ids_0=input_ids_with_markers
                    )

                task_encodings.append(
                    TaskEncoding(
                        document=document,
                        inputs={"input_ids": input_ids_with_markers},
                        metadata={"candidate_annotation": rel},
                    )
                )

        return task_encodings

    def _maybe_log_example(
        self,
        task_encoding: TaskEncodingType,
        target: TargetEncodingType,
    ):
        """Maybe log the example."""

        # log the first n examples
        if self._logged_examples_counter < self.log_first_n_examples:
            input_ids = task_encoding.inputs["input_ids"]
            tokens = self.tokenizer.convert_ids_to_tokens(input_ids)
            target_labels = [self.id_to_label[label_id] for label_id in target]
            logger.info("*** Example ***")
            logger.info("doc id: %s", task_encoding.document.id)
            logger.info("tokens: %s", " ".join([str(x) for x in tokens]))
            logger.info("input_ids: %s", " ".join([str(x) for x in input_ids]))
            logger.info("Expected label: %s (ids = %s)", target_labels, target)

            self._logged_examples_counter += 1

    def encode_target(
        self,
        task_encoding: TaskEncodingType,
    ) -> TargetEncodingType:
        candidate_annotation = task_encoding.metadata["candidate_annotation"]
        if isinstance(candidate_annotation, BinaryRelation):
            labels = [candidate_annotation.label]
        else:
            raise NotImplementedError(
                f"encoding the target with a candidate_annotation of another type than BinaryRelation is "
                f"not yet supported. candidate_annotation has the type: {type(candidate_annotation)}"
            )
        target = [self.label_to_id[label] for label in labels]

        self._maybe_log_example(task_encoding=task_encoding, target=target)

        return target

    def unbatch_output(self, model_output: ModelOutputType) -> Sequence[TaskOutputType]:
        logits = model_output["logits"]

        output_label_probs = logits.sigmoid() if self.multi_label else logits.softmax(dim=-1)
        output_label_probs = output_label_probs.detach().cpu().numpy()

        unbatched_output = []
        if self.multi_label:
            raise NotImplementedError
        else:
            label_ids = np.argmax(output_label_probs, axis=-1)
            for batch_idx, label_id in enumerate(label_ids):
                label = self.id_to_label[label_id]
                prob = float(output_label_probs[batch_idx, label_id])
                result: TaskOutputType = {
                    "labels": [label],
                    "probabilities": [prob],
                }
                unbatched_output.append(result)

        return unbatched_output

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncodingType,
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, Union[BinaryRelation, MultiLabeledBinaryRelation, NaryRelation]]]:
        candidate_annotation = task_encoding.metadata["candidate_annotation"]
        new_annotation: Union[BinaryRelation, MultiLabeledBinaryRelation, NaryRelation]
        if self.multi_label:
            raise NotImplementedError
        else:
            label = task_output["labels"][0]
            probability = task_output["probabilities"][0]
            if isinstance(candidate_annotation, BinaryRelation):
                head = candidate_annotation.head
                tail = candidate_annotation.tail
                # reverse any predicted reversed relations back
                if self.reversed_relation_label_suffix is not None and label.endswith(
                    self.reversed_relation_label_suffix
                ):
                    label = label[: -len(self.reversed_relation_label_suffix)]
                    head, tail = tail, head
                new_annotation = BinaryRelation(
                    head=head, tail=tail, label=label, score=probability
                )
            elif isinstance(candidate_annotation, NaryRelation):
                if self.reversed_relation_label_suffix is not None:
                    raise ValueError(f"can not reverse a NaryRelation")
                new_annotation = NaryRelation(
                    arguments=candidate_annotation.arguments,
                    roles=candidate_annotation.roles,
                    label=label,
                    score=probability,
                )
            else:
                raise NotImplementedError(
                    f"creating a new annotation from a candidate_annotation of another type than BinaryRelation is "
                    f"not yet supported. candidate_annotation has the type: {type(candidate_annotation)}"
                )
            if label != self.none_label:
                yield self.relation_annotation, new_annotation

    def collate(self, task_encodings: Sequence[TaskEncodingType]) -> ModelStepInputType:
        input_features = [task_encoding.inputs for task_encoding in task_encodings]

        inputs: Dict[str, torch.Tensor] = self.tokenizer.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        if not task_encodings[0].has_targets:
            return inputs, None

        target_list: List[TargetEncodingType] = [
            task_encoding.targets for task_encoding in task_encodings
        ]
        targets = torch.tensor(target_list, dtype=torch.int64)

        if not self.multi_label:
            targets = targets.flatten()

        return inputs, targets



from abc import ABC, abstractmethod
from typing import Dict, Generic, Iterable, Optional, TypeVar, Union

from pytorch_ie.core.document import Document
from pytorch_ie.core.module_mixins import WithDocumentTypeMixin

T = TypeVar("T")


class DocumentMetric(ABC, WithDocumentTypeMixin, Generic[T]):
    """This defines the interface for a document metric."""

    def __init__(self):
        self.reset()
        self._current_split: Optional[str] = None

    @abstractmethod
    def reset(self) -> None:
        """Any reset logic that needs to be performed before the metric is called again."""

    def __call__(
        self,
        document_or_collection: Union[Iterable[Document], Document, Dict[str, Iterable[Document]]],
    ) -> Union[Dict[str, T], T]:
        """This method is called to update the metric with a document or collection of documents.

        If a collection is passed, the metric is also computed and the result is returned. If the
        collection is a dictionary, the metric is computed for each split and the result is
        returned as a dictionary.
        """
        if isinstance(document_or_collection, Document):
            # do not reset here to allow for multiple calls
            self._update(document_or_collection)
            return self.compute(reset=False)
        elif isinstance(document_or_collection, dict):
            result: Dict[str, T] = {}
            for split_name, split in document_or_collection.items():
                self._current_split = split_name
                self.reset()
                split_values: T = self(split)  # type: ignore
                result[split_name] = split_values
                self._current_split = None
            return result
        elif isinstance(document_or_collection, Iterable):
            for doc in document_or_collection:
                if not isinstance(doc, Document):
                    raise TypeError(
                        f"document_or_collection contains an object that is not a document: {type(doc)}"
                    )
                self._update(doc)
            # do not reset here to allow for multiple calls
            return self.compute(reset=False)
        else:
            raise TypeError(
                f"document_or_collection has unknown type: {type(document_or_collection)}"
            )

    def compute(self, reset: bool = True) -> T:
        metric_values = self._compute()
        if reset:
            self.reset()
        return metric_values

    @abstractmethod
    def _update(self, document: Document) -> None:
        """This method is called to update the metric with the new document."""

    @abstractmethod
    def _compute(self) -> T:
        """This method is called to get the metric values."""

    @property
    def current_split(self) -> Optional[str]:
        """The current split that is being processed."""
        return self._current_split



# taken from hydra/_internal/instantiate/_instantiate2.py
from typing import Any, Callable, Optional, Type, Union

from pytorch_ie.core.document import Document


class HydraException(Exception):
    ...


class CompactHydraException(HydraException):
    ...


class InstantiationException(CompactHydraException):
    ...


def _locate(path: str) -> Any:
    """Locate an object by name or dotted path, importing as necessary.

    This is similar to the pydoc function `locate`, except that it checks for
    the module from the given path from back to front.
    """
    if path == "":
        raise ImportError("Empty path")
    from importlib import import_module
    from types import ModuleType

    parts = [part for part in path.split(".")]
    for part in parts:
        if not len(part):
            raise ValueError(
                f"Error loading '{path}': invalid dotstring."
                + "\nRelative imports are not supported."
            )
    assert len(parts) > 0
    part0 = parts[0]
    try:
        obj = import_module(part0)
    except Exception as exc_import:
        raise ImportError(
            f"Error loading '{path}':\n{repr(exc_import)}"
            + f"\nAre you sure that module '{part0}' is installed?"
        ) from exc_import
    for m in range(1, len(parts)):
        part = parts[m]
        try:
            obj = getattr(obj, part)
        except AttributeError as exc_attr:
            parent_dotpath = ".".join(parts[:m])
            if isinstance(obj, ModuleType):
                mod = ".".join(parts[: m + 1])
                try:
                    obj = import_module(mod)
                    continue
                except ModuleNotFoundError as exc_import:
                    raise ImportError(
                        f"Error loading '{path}':\n{repr(exc_import)}"
                        + f"\nAre you sure that '{part}' is importable from module '{parent_dotpath}'?"
                    ) from exc_import
                except Exception as exc_import:
                    raise ImportError(
                        f"Error loading '{path}':\n{repr(exc_import)}"
                    ) from exc_import
            raise ImportError(
                f"Error loading '{path}':\n{repr(exc_attr)}"
                + f"\nAre you sure that '{part}' is an attribute of '{parent_dotpath}'?"
            ) from exc_attr
    return obj


def resolve_target(
    target: Union[str, type, Callable[..., Any]], full_key: str = ""
) -> Union[type, Callable[..., Any]]:
    """Resolve target string, type or callable into type or callable."""
    if isinstance(target, str):
        try:
            target = _locate(target)
        except Exception as e:
            msg = f"Error locating target '{target}', set env var HYDRA_FULL_ERROR=1 to see chained exception."
            if full_key:
                msg += f"\nfull_key: {full_key}"
            raise InstantiationException(msg) from e
    if not callable(target):
        msg = f"Expected a callable target, got '{target}' of type '{type(target).__name__}'"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise InstantiationException(msg)
    return target


def resolve_optional_document_type(
    document_type: Optional[Union[str, Type[Document]]]
) -> Optional[Type[Document]]:
    if document_type is None:
        return None
    if isinstance(document_type, str):
        dt = resolve_target(document_type)
    else:
        dt = document_type
    if not (isinstance(dt, type) and issubclass(dt, Document)):
        raise TypeError(
            f"(resolved) document_type must be a subclass of Document, but it is: {dt}"
        )
    return dt


def serialize_document_type(document_type: Type[Document]) -> str:
    return f"{document_type.__module__}.{document_type.__name__}"



from typing import Any, Dict, Optional, Tuple

import torch
import torchmetrics
from torch import Tensor, nn
from transformers import AutoConfig, AutoModelForTokenClassification, BatchEncoding
from typing_extensions import TypeAlias

from pytorch_ie.core import PyTorchIEModel
from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses

ModelInputType: TypeAlias = BatchEncoding
ModelOutputType: TypeAlias = Dict[str, Any]

ModelStepInputType: TypeAlias = Tuple[
    ModelInputType,
    Optional[Tensor],
]


TRAINING = "train"
VALIDATION = "val"
TEST = "test"


@PyTorchIEModel.register()
class TransformerTokenClassificationModel(
    PyTorchIEModel, RequiresModelNameOrPath, RequiresNumClasses
):
    def __init__(
        self,
        model_name_or_path: str,
        num_classes: int,
        learning_rate: float = 1e-5,
        label_pad_token_id: int = -100,
        ignore_index: int = 0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.learning_rate = learning_rate
        self.label_pad_token_id = label_pad_token_id
        self.num_classes = num_classes

        config = AutoConfig.from_pretrained(model_name_or_path, num_labels=num_classes)
        if self.is_from_pretrained:
            self.model = AutoModelForTokenClassification.from_config(config=config)
        else:
            self.model = AutoModelForTokenClassification.from_pretrained(
                model_name_or_path, config=config
            )

        self.f1 = nn.ModuleDict(
            {
                f"stage_{stage}": torchmetrics.F1Score(
                    num_classes=num_classes, ignore_index=ignore_index, task="multiclass"
                )
                for stage in [TRAINING, VALIDATION, TEST]
            }
        )

    def forward(self, inputs: ModelInputType) -> ModelOutputType:
        return self.model(**inputs)

    def step(
        self,
        stage: str,
        batch: ModelStepInputType,
    ):
        inputs, target = batch
        assert target is not None, "target has to be available for training"

        inputs["labels"] = target
        output = self(inputs)

        loss = output.loss
        # show loss on each step only during training
        self.log(f"{stage}/loss", loss, on_step=(stage == TRAINING), on_epoch=True, prog_bar=True)

        target_flat = target.view(-1)

        valid_indices = target_flat != self.label_pad_token_id
        valid_logits = output.logits.view(-1, self.num_classes)[valid_indices]
        valid_target = target_flat[valid_indices]

        f1 = self.f1[f"stage_{stage}"]
        f1(valid_logits, valid_target)
        self.log(f"{stage}/f1", f1, on_step=False, on_epoch=True, prog_bar=True)

        return loss

    def training_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TRAINING, batch=batch)

    def validation_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=VALIDATION, batch=batch)

    def test_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TEST, batch=batch)

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.learning_rate)



import functools
import logging
from typing import (
    Callable,
    Counter,
    DefaultDict,
    Dict,
    List,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
)

from transformers import PreTrainedTokenizer

from pytorch_ie.annotations import LabeledSpan, Span

TypedSpan = Tuple[int, Tuple[int, int]]
TypedStringSpan = Tuple[str, Tuple[int, int]]

logger = logging.getLogger(__name__)


class InvalidTagSequence(Exception):
    def __init__(self, tag_sequence=None):
        super().__init__()
        self.tag_sequence = tag_sequence

    def __str__(self):
        return " ".join(self.tag_sequence)


def bio_tags_to_spans(
    tag_sequence: Sequence[str],
    classes_to_ignore: Optional[List[str]] = None,
    include_ill_formed: bool = True,
) -> List[TypedStringSpan]:
    """
    Given a sequence corresponding to BIO tags, extracts spans.
    Spans are inclusive and can be of zero length, representing a single word span.
    Ill-formed spans are also included (i.e those which do not start with a "B-LABEL"),
    as otherwise it is possible to get a perfect precision score whilst still predicting
    ill-formed spans in addition to the correct spans. This function works properly when
    the spans are unlabeled (i.e., your labels are simply "B", "I", and "O").
    # Parameters
    tag_sequence : `List[str]`, required.
        The integer class labels for a sequence.
    classes_to_ignore : `List[str]`, optional (default = `None`).
        A list of string class labels `excluding` the bio tag
        which should be ignored when extracting spans.
    include_ill_formed: `bool`, optional (default = `True`).
        If this flag is enabled, include spans that do not start with "B". Otherwise, these are ignored.
    # Returns
    spans : `List[TypedStringSpan]`
        The typed, extracted spans from the sequence, in the format (label, (span_start, span_end)).
        Note that the label `does not` contain any BIO tag prefixes.
    """
    classes_to_ignore = classes_to_ignore or []
    spans: Set[Tuple[str, Tuple[int, int]]] = set()
    span_start = 0
    span_end = 0
    active_conll_tag = None
    for index, string_tag in enumerate(tag_sequence):
        # Actual BIO tag.
        bio_tag = string_tag[0]
        if bio_tag not in ["B", "I", "O"]:
            raise InvalidTagSequence(tag_sequence)
        conll_tag = string_tag[2:]
        if bio_tag == "O" or conll_tag in classes_to_ignore:
            # The span has ended.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            active_conll_tag = None
            # We don't care about tags we are
            # told to ignore, so we do nothing.
            continue
        elif bio_tag == "B":
            # We are entering a new span; reset indices
            # and active tag to new span.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            active_conll_tag = conll_tag
            span_start = index
            span_end = index
        elif bio_tag == "I" and conll_tag == active_conll_tag:
            # We're inside a span.
            span_end += 1
        else:
            # This is the case the bio label is an "I", but either:
            # 1) the span hasn't started - i.e. an ill formed span.
            # 2) The span is an I tag for a different conll annotation.
            # We'll process the previous span if it exists, but also
            # include this span. This is important, because otherwise,
            # a model may get a perfect F1 score whilst still including
            # false positive ill-formed spans.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            if include_ill_formed:
                active_conll_tag = conll_tag
                span_start = index
                span_end = index
            else:
                active_conll_tag = None
                # We don't care about tags we are
                # told to ignore, so we do nothing.
                continue
    # Last token might have been a part of a valid span.
    if active_conll_tag is not None:
        spans.add((active_conll_tag, (span_start, span_end)))
    return list(spans)


def io_tags_to_spans(
    tag_sequence: List[str], classes_to_ignore: Optional[List[str]] = None
) -> List[TypedStringSpan]:
    """
    Decode spans from simple IO encoding tag sequence, i.e. tags with an expected tag set of labels + "O".
    Create spans from maximal subsequences that have the same label.

    # Parameters
    tag_sequence : `List[str]`, required.
        The integer class labels for a sequence.
    classes_to_ignore : `List[str]`, optional (default = `None`).
        A list of string class labels `excluding` the bio tag
        which should be ignored when extracting spans.
    # Returns
    spans : `List[TypedStringSpan]`
        The typed, extracted spans from the sequence, in the format (label, (span_start, span_end)).
        Note that the label `does not` contain any BIO tag prefixes.
    """
    classes_to_ignore = classes_to_ignore or []
    spans: Set[Tuple[str, Tuple[int, int]]] = set()
    span_start = 0
    span_end = 0
    active_tag = None
    for index, string_tag in enumerate(tag_sequence):
        if string_tag == "O" or string_tag in classes_to_ignore:
            # The span has ended.
            if active_tag is not None:
                spans.add((active_tag, (span_start, span_end)))
            active_tag = None
            # We don't care about tags we are
            # told to ignore, so we do nothing.
            continue
        elif string_tag == active_tag:
            # We're inside a span.
            span_end += 1
        else:
            if active_tag is not None:
                spans.add((active_tag, (span_start, span_end)))
            active_tag = string_tag
            span_start = index
            span_end = index
    # Last token might have been a part of a valid span.
    if active_tag is not None:
        spans.add((active_tag, (span_start, span_end)))
    return list(spans)


def convert_span_annotations_to_tag_sequence(
    spans: Sequence[LabeledSpan],
    special_tokens_mask: Sequence[int],
    char_to_token_mapper: Callable[[int], Optional[int]],
    partition: Optional[Span] = None,
    statistics: Optional[DefaultDict[str, Counter]] = None,
) -> Optional[MutableSequence[Optional[str]]]:
    """
    Given a list of span annotations, a character position to token mapper (as obtained from
    batch_encoding.char_to_token) and a special tokens mask, create a sequence of tags with the length of the
    special tokens mask. For special token positions, None is returned as tag.
    If a partition is provided, only the tokens within that span are considered.
    For now, the BIO-encoding is used.
    Note: The spans are not allowed to overlap (will raise an exception).
    """
    tag_sequence = [
        None if special_tokens_mask[j] else "O" for j in range(len(special_tokens_mask))
    ]
    offset = partition.start if partition is not None else 0
    for span in spans:
        if partition is not None and (span.start < partition.start or span.end > partition.end):
            continue

        start_idx = char_to_token_mapper(span.start - offset)
        end_idx = char_to_token_mapper(span.end - 1 - offset)
        if start_idx is None or end_idx is None:
            if statistics is not None:
                statistics["skipped_unaligned"][span.label] += 1
            else:
                logger.warning(
                    f"Entity annotation does not start or end with a token, it will be skipped: {span}"
                )
            continue

        # negative numbers encode out-of-window tokens
        if start_idx < 0 or end_idx < 0:
            continue

        for j in range(start_idx, end_idx + 1):
            if tag_sequence[j] is not None and tag_sequence[j] != "O":
                logger.warning(f"tag already assigned (current span has an overlap: {span}).")
                return None
            prefix = "B" if j == start_idx else "I"
            tag_sequence[j] = f"{prefix}-{span.label}"

        if statistics is not None:
            statistics["added"][span.label] += 1

    return tag_sequence


def get_token_slice(
    character_slice: Tuple[int, int],
    char_to_token_mapper: Callable[[int], Optional[int]],
    character_offset: int = 0,
) -> Optional[Tuple[int, int]]:
    """
    Using an encoding to map a character slice to the respective token slice. If the slice start or end does
    not match a token start or end respectively, return None.
    """
    start = char_to_token_mapper(character_slice[0] - character_offset)
    before_end = char_to_token_mapper(character_slice[1] - 1 - character_offset)
    if start is None or before_end is None:
        return None
    return start, before_end + 1


def is_contained_in(start_end: Tuple[int, int], other_start_end: Tuple[int, int]) -> bool:
    return other_start_end[0] <= start_end[0] and start_end[1] <= other_start_end[1]


def has_overlap(start_end: Tuple[int, int], other_start_end: Tuple[int, int]):
    return (
        start_end[0] <= other_start_end[0] < start_end[1]
        or start_end[0] < other_start_end[1] <= start_end[1]
        or other_start_end[0] <= start_end[0] < other_start_end[1]
        or other_start_end[0] < start_end[1] <= other_start_end[1]
    )


def _char_to_token_mapper(
    char_idx: int,
    char_to_token_mapping: Dict[int, int],
    char_start: Optional[int] = None,
    char_end: Optional[int] = None,
) -> Optional[int]:
    if char_start is not None and char_idx < char_start:
        # return negative number to encode out-ot-window
        return -1
    if char_end is not None and char_idx >= char_end:
        # return negative number to encode out-ot-window
        return -2
    return char_to_token_mapping.get(char_idx, None)


def get_char_to_token_mapper(
    char_to_token_mapping: Dict[int, int],
    char_start: Optional[int] = None,
    char_end: Optional[int] = None,
) -> Callable[[int], Optional[int]]:
    return functools.partial(
        _char_to_token_mapper,
        char_to_token_mapping=char_to_token_mapping,
        char_start=char_start,
        char_end=char_end,
    )


def get_special_token_mask(token_ids_0: List[int], tokenizer: PreTrainedTokenizer) -> List[int]:
    # TODO: check why we can not just use tokenizer.get_special_tokens_mask()
    #  (this checks if token_ids_1 is not None and raises an exception)

    # exclude unknown token id since this indicate a real input token
    special_ids = set(tokenizer.all_special_ids) - {tokenizer.unk_token_id}
    return [1 if token_id in special_ids else 0 for token_id in token_ids_0]


def tokens_and_tags_to_text_and_labeled_spans(
    tokens: Sequence[str], tags: Sequence[str]
) -> Tuple[str, Sequence[LabeledSpan]]:
    start = 0
    token_offsets: List[Tuple[int, int]] = []
    for token in tokens:
        end = start + len(token)
        token_offsets.append((start, end))
        # we add a space after each token
        start = end + 1

    text = " ".join(tokens)

    spans: List[LabeledSpan] = []
    for label, (start, end) in bio_tags_to_spans(tags):
        spans.append(
            LabeledSpan(start=token_offsets[start][0], end=token_offsets[end][1], label=label)
        )

    return text, spans



from dataclasses import dataclass, field
from typing import Optional, Tuple

from pytorch_ie.core.document import Annotation


def _post_init_single_label(self):
    if not isinstance(self.label, str):
        raise ValueError("label must be a single string.")

    if not isinstance(self.score, float):
        raise ValueError("score must be a single float.")


def _post_init_multi_label(self):
    if self.score is None:
        score = tuple([1.0] * len(self.label))
        object.__setattr__(self, "score", score)

    if not isinstance(self.label, tuple):
        object.__setattr__(self, "label", tuple(self.label))

    if not isinstance(self.score, tuple):
        object.__setattr__(self, "score", tuple(self.score))

    if len(self.label) != len(self.score):
        raise ValueError(
            f"Number of labels ({len(self.label)}) and scores ({len(self.score)}) must be equal."
        )


def _post_init_multi_span(self):
    if isinstance(self.slices, list):
        object.__setattr__(self, "slices", tuple(tuple(s) for s in self.slices))


def _post_init_arguments_and_roles(self):
    if len(self.arguments) != len(self.roles):
        raise ValueError(
            f"Number of arguments ({len(self.arguments)}) and roles ({len(self.roles)}) must be equal"
        )
    if not isinstance(self.arguments, tuple):
        object.__setattr__(self, "arguments", tuple(self.arguments))
    if not isinstance(self.roles, tuple):
        object.__setattr__(self, "roles", tuple(self.roles))


@dataclass(eq=True, frozen=True)
class Label(Annotation):
    label: str
    score: float = field(default=1.0, compare=False)

    def __post_init__(self) -> None:
        _post_init_single_label(self)


@dataclass(eq=True, frozen=True)
class MultiLabel(Annotation):
    label: Tuple[str, ...]
    score: Optional[Tuple[float, ...]] = field(default=None, compare=False)

    def __post_init__(self) -> None:
        _post_init_multi_label(self)


@dataclass(eq=True, frozen=True)
class Span(Annotation):
    start: int
    end: int

    def __str__(self) -> str:
        if self.target is None:
            return ""
        return str(self.target[self.start : self.end])


@dataclass(eq=True, frozen=True)
class LabeledSpan(Span):
    label: str
    score: float = field(default=1.0, compare=False)

    def __post_init__(self) -> None:
        _post_init_single_label(self)


@dataclass(eq=True, frozen=True)
class MultiLabeledSpan(Span):
    label: Tuple[str, ...]
    score: Optional[Tuple[float, ...]] = field(default=None, compare=False)

    def __post_init__(self) -> None:
        _post_init_multi_label(self)


@dataclass(eq=True, frozen=True)
class LabeledMultiSpan(Annotation):
    slices: Tuple[Tuple[int, int], ...]
    label: str
    score: float = field(default=1.0, compare=False)

    def __post_init__(self) -> None:
        _post_init_multi_span(self)
        _post_init_single_label(self)


@dataclass(eq=True, frozen=True)
class MultiLabeledMultiSpan(Annotation):
    slices: Tuple[Tuple[int, int], ...]
    label: Tuple[str, ...]
    score: Optional[Tuple[float, ...]] = field(default=None, compare=False)

    def __post_init__(self) -> None:
        _post_init_multi_span(self)
        _post_init_multi_label(self)


@dataclass(eq=True, frozen=True)
class BinaryRelation(Annotation):
    head: Annotation
    tail: Annotation
    label: str
    score: float = field(default=1.0, compare=False)

    def __post_init__(self) -> None:
        _post_init_single_label(self)


@dataclass(eq=True, frozen=True)
class MultiLabeledBinaryRelation(Annotation):
    head: Annotation
    tail: Annotation
    label: Tuple[str, ...]
    score: Optional[Tuple[float, ...]] = field(default=None, compare=False)

    def __post_init__(self) -> None:
        _post_init_multi_label(self)


@dataclass(eq=True, frozen=True)
class NaryRelation(Annotation):
    arguments: Tuple[Annotation, ...]
    roles: Tuple[str, ...]
    label: str
    score: float = field(default=1.0, compare=False)

    def __post_init__(self) -> None:
        _post_init_arguments_and_roles(self)
        _post_init_single_label(self)



import logging
from abc import abstractmethod
from collections import defaultdict
from typing import Any, Dict, Generator, List, Optional, Tuple, Type, Union

from pytorch_ie.core.document import Document
from pytorch_ie.core.metric import DocumentMetric
from pytorch_ie.utils.hydra import (
    InstantiationException,
    resolve_optional_document_type,
    resolve_target,
)

logger = logging.getLogger(__name__)


def _flatten_dict_gen(d, parent_key: Tuple[str, ...] = ()) -> Generator:
    for k, v in d.items():
        new_key = parent_key + (k,)
        if isinstance(v, dict):
            yield from dict(_flatten_dict_gen(v, new_key)).items()
        else:
            yield new_key, v


def flatten_dict(d: Dict[str, Any]) -> Dict[Tuple[str, ...], Any]:
    return dict(_flatten_dict_gen(d))


def unflatten_dict(d: Dict[Tuple[str, ...], Any]) -> Union[Dict[str, Any], Any]:
    """Unflattens a dictionary with nested keys.

    Example:
        >>> d = {("a", "b", "c"): 1, ("a", "b", "d"): 2, ("a", "e"): 3}
        >>> unflatten_dict(d)
        {'a': {'b': {'c': 1, 'd': 2}, 'e': 3}}
    """
    result: Dict[str, Any] = {}
    for k, v in d.items():
        if len(k) == 0:
            if len(result) > 1:
                raise ValueError("Cannot unflatten dictionary with multiple root keys.")
            return v
        current = result
        for key in k[:-1]:
            current = current.setdefault(key, {})
        current[k[-1]] = v
    return result


def _min(values: List[float]) -> Optional[float]:
    if len(values) == 0:
        return None
    return min(values)


def _max(values: List[float]) -> Optional[float]:
    if len(values) == 0:
        return None
    return max(values)


def _mean(values: List[float]) -> Optional[float]:
    if len(values) == 0:
        return None
    return sum(values) / len(values)


def _median(values: List[float]) -> Optional[float]:
    if len(values) == 0:
        return None
    return sorted(values)[len(values) // 2]


def _std(values: List[float]) -> Optional[float]:
    mean_value = _mean(values)
    if mean_value is None:
        return None
    return (sum((x - mean_value) ** 2 for x in values) / len(values)) ** 0.5


AGGREGATION_FUNCTIONS = {
    "min": _min,
    "max": _max,
    "mean": _mean,
    "median": _median,
    "std": _std,
}


def resolve_agg_function(name: str):
    if name in AGGREGATION_FUNCTIONS:
        return AGGREGATION_FUNCTIONS[name]
    else:
        try:
            return resolve_target(name)
        except InstantiationException:
            try:
                return resolve_target(f"builtins.{name}")
            except InstantiationException:
                raise ImportError(f"Cannot resolve aggregation function: {name}")


class DocumentStatistic(DocumentMetric):
    """A special type of metric that collects statistics from a document.

    Usage:

    ```python
    from transformers import AutoTokenizer, PreTrainedTokenizer
    from pytorch_ie import DatasetDict
    from pytorch_ie.core import Document, DocumentStatistic

    class TokenCountCollector(DocumentStatistic):

        def __init__(
            self,
            tokenizer: Union[str, PreTrainedTokenizer],
            text_field: str,
            tokenizer_kwargs: Optional[Dict[str, Any]] = None,
            **kwargs,
        ):
            super().__init__(**kwargs)
            self.tokenizer = (
                AutoTokenizer.from_pretrained(tokenizer) if isinstance(tokenizer, str) else tokenizer
            )
            self.tokenizer_kwargs = tokenizer_kwargs or {}
            self.text_field = text_field

        def _collect(self, doc: Document) -> int:
            text = getattr(doc, self.text_field)
            encodings = self.tokenizer(text, **self.tokenizer_kwargs)
            tokens = encodings.tokens()
            return len(tokens)

    dataset = DatasetDict.load_dataset("pie/conll2003")
    statistic = TokenCountCollector(
        text_field="text",
        tokenizer="bert-base-uncased",
        tokenizer_kwargs=dict(add_special_tokens=False),
    )
    values = statistic(dataset)
    assert values == {
        'train': {'mean': 17.950502100989958, 'std': 13.016237876955675, 'min': 1, 'max': 162},
        'validation': {'mean': 19.368307692307692, 'std': 14.583363922289669, 'min': 1, 'max': 144},
        'test': {'mean': 16.774978279756734, 'std': 13.176981022988947, 'min': 1, 'max': 138}
    }
    ```
    """

    DEFAULT_AGGREGATION_FUNCTIONS = ["mean", "std", "min", "max"]

    def __init__(
        self,
        show_histogram: bool = False,
        show_as_markdown: bool = False,
        aggregation_functions: Optional[List[str]] = None,
        title: Optional[str] = None,
        document_type: Optional[Union[Type[Document], str]] = None,
    ) -> None:
        super().__init__()
        self.aggregation_functions = {
            f_name: resolve_agg_function(f_name)
            for f_name in aggregation_functions or self.DEFAULT_AGGREGATION_FUNCTIONS
        }
        self.show_histogram = show_histogram
        self.show_as_markdown = show_as_markdown
        self.title = title or self.__class__.__name__
        self._document_type = resolve_optional_document_type(document_type)

    @property
    def document_type(self) -> Optional[Type[Document]]:
        return self._document_type or super().document_type

    def reset(self) -> None:
        self._values: List[Any] = []

    @abstractmethod
    def _collect(self, doc: Document) -> Any:
        """Collect any values from a document."""

    def _update(self, document: Document) -> None:
        values = self._collect(document)
        self._values.append(values)

    def _compute(self) -> Any:
        """We just integrate the values by creating lists for each leaf of the (nested)
        dictionary."""
        stats = defaultdict(list)
        for collected_result in self._values:
            if isinstance(collected_result, dict):
                collected_result_flat = flatten_dict(collected_result)
                for k, v in collected_result_flat.items():
                    if isinstance(v, list):
                        stats[k].extend(v)
                    else:
                        stats[k].append(v)
            else:
                if isinstance(collected_result, list):
                    stats[()].extend(collected_result)
                else:
                    stats[()].append(collected_result)
        if self.current_split is not None:
            title = f"{self.title} (split: {self.current_split}, {len(self._values)} documents)"
        else:
            title = f"{self.title} ({len(self._values)} documents)"
        if self.show_histogram:
            import plotext as plt

            for k, values in stats.items():
                if isinstance(values, list):
                    plt.hist(values, label=".".join(k) if len(k) > 0 else None)
            plt.title(title)
            plt.show()
            plt.clear_figure()

        aggregated_stats = {}
        for k, v in stats.items():
            for f_name, f in self.aggregation_functions.items():
                aggregated_stats[k + (f_name,)] = f(v)

        if self.show_as_markdown:
            import pandas as pd

            series = pd.Series(aggregated_stats)
            if len(series.index.levels) > 1:
                df = series.unstack(-1)
                logger.info(f"{title}\n{df.round(3).to_markdown()}")
            else:
                series.index = series.index.get_level_values(0)
                logger.info(f"{title}\n{series.round(3).to_markdown()}")

        return unflatten_dict(aggregated_stats)



import os
from pathlib import Path
from typing import Any, Dict, Optional, Type, Union

import torch
from huggingface_hub.constants import PYTORCH_WEIGHTS_NAME
from huggingface_hub.file_download import hf_hub_download

from pytorch_ie.core import PyTorchIEModel, TaskModule
from pytorch_ie.core.hf_hub_mixin import PieModelHFHubMixin, PieTaskModuleHFHubMixin
from pytorch_ie.pipeline import Pipeline


class AutoModel(PieModelHFHubMixin):
    @classmethod
    def _from_pretrained(
        cls,
        *,
        model_id: str,
        revision: Optional[str],
        cache_dir: Optional[Union[str, Path]],
        force_download: bool,
        proxies: Optional[Dict],
        resume_download: bool,
        local_files_only: bool,
        token: Union[str, bool, None],
        map_location: str = "cpu",
        strict: bool = False,
        config: Optional[dict] = None,
        **model_kwargs,
    ) -> PyTorchIEModel:
        """
        Overwrite this method in case you wish to initialize your model in a different way.
        """

        """Load Pytorch pretrained weights and return the loaded model."""
        if os.path.isdir(model_id):
            print("Loading weights from local directory")
            model_file = os.path.join(model_id, PYTORCH_WEIGHTS_NAME)
        else:
            model_file = hf_hub_download(
                repo_id=model_id,
                filename=PYTORCH_WEIGHTS_NAME,
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                token=token,
                local_files_only=local_files_only,
            )

        config = (config or {}).copy()
        config.update(model_kwargs)
        class_name = config.pop(cls.config_type_key)
        clazz = PyTorchIEModel.by_name(class_name)
        model = clazz(**config)

        state_dict = torch.load(model_file, map_location=torch.device(map_location))
        model.load_state_dict(state_dict, strict=strict)  # type: ignore
        model.eval()  # type: ignore

        return model

    @classmethod
    def from_config(cls, config: dict, **kwargs) -> PyTorchIEModel:
        """Build a model from a config dict."""
        config = config.copy()
        class_name = config.pop(cls.config_type_key)
        clazz = PyTorchIEModel.by_name(class_name)
        return clazz._from_config(config, **kwargs)


class AutoTaskModule(PieTaskModuleHFHubMixin):
    @classmethod
    def _from_pretrained(  # type: ignore
        cls,
        *,
        model_id: str,
        revision: Optional[str],
        cache_dir: Optional[Union[str, Path]],
        force_download: bool,
        proxies: Optional[Dict],
        resume_download: bool,
        local_files_only: bool,
        token: Union[str, bool, None],
        map_location: str = "cpu",
        strict: bool = False,
        config: Optional[dict] = None,
        **taskmodule_kwargs,
    ) -> TaskModule:
        config = (config or {}).copy()
        config.update(taskmodule_kwargs)
        class_name = config.pop(cls.config_type_key)
        clazz: Type[TaskModule] = TaskModule.by_name(class_name)
        taskmodule = clazz(**config)
        taskmodule.post_prepare()
        return taskmodule

    @classmethod
    def from_config(cls, config: dict, **kwargs) -> TaskModule:
        """Build a task module from a config dict."""
        config = config.copy()
        class_name = config.pop(cls.config_type_key)
        clazz: Type[TaskModule] = TaskModule.by_name(class_name)
        return clazz._from_config(config, **kwargs)


class AutoPipeline:
    @staticmethod
    def from_pretrained(
        pretrained_model_name_or_path: str,
        force_download: bool = False,
        resume_download: bool = False,
        proxies: Optional[Dict] = None,
        use_auth_token: Optional[str] = None,
        cache_dir: Optional[str] = None,
        local_files_only: bool = False,
        taskmodule_kwargs: Optional[Dict[str, Any]] = None,
        model_kwargs: Optional[Dict[str, Any]] = None,
        device: int = -1,
        binary_output: bool = False,
        **kwargs,
    ) -> Pipeline:
        taskmodule_kwargs = taskmodule_kwargs or {}
        model_kwargs = model_kwargs or {}

        taskmodule = AutoTaskModule.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            force_download=force_download,
            resume_download=resume_download,
            proxies=proxies,
            use_auth_token=use_auth_token,
            cache_dir=cache_dir,
            local_files_only=local_files_only,
            **taskmodule_kwargs,
        )

        model = AutoModel.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            force_download=force_download,
            resume_download=resume_download,
            proxies=proxies,
            use_auth_token=use_auth_token,
            cache_dir=cache_dir,
            local_files_only=local_files_only,
            **model_kwargs,
        )

        return Pipeline(
            taskmodule=taskmodule,
            model=model,
            device=device,
            binary_output=binary_output,
            **kwargs,
        )



from typing import Any, Dict

from pytorch_lightning import LightningModule

from pytorch_ie.core.hf_hub_mixin import PieModelHFHubMixin
from pytorch_ie.core.registrable import Registrable


class PyTorchIEModel(PieModelHFHubMixin, LightningModule, Registrable):
    def _config(self) -> Dict[str, Any]:
        config = super()._config() or {}
        config[self.config_type_key] = PyTorchIEModel.name_for_object_class(self)
        # add all hparams
        config.update(self.hparams)
        return config

    def predict(
        self,
        inputs: Any,
        **kwargs,
    ) -> Any:
        return self(inputs, **kwargs)



"""
workflow:
    document
        -> (input_encoding, target_encoding) -> task_encoding
            -> model_encoding -> model_output
        -> task_output
    -> document
"""

import logging
from typing import Any, Dict, Iterator, MutableMapping, Optional, Sequence, Tuple, TypedDict, Union

import numpy as np
import torch
from transformers import AutoTokenizer
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import TruncationStrategy
from typing_extensions import TypeAlias

from pytorch_ie.annotations import Label
from pytorch_ie.core import TaskEncoding, TaskModule
from pytorch_ie.documents import TextDocumentWithLabel
from pytorch_ie.models.transformer_text_classification import ModelOutputType, ModelStepInputType

logger = logging.getLogger(__name__)


class TaskOutput(TypedDict, total=False):
    label: str
    probability: float


# Define task specific input and output types
DocumentType: TypeAlias = TextDocumentWithLabel
InputEncodingType: TypeAlias = MutableMapping[str, Any]
TargetEncodingType: TypeAlias = int
ModelEncodingType: TypeAlias = ModelStepInputType
ModelOutputType: TypeAlias = ModelOutputType
TaskOutputType: TypeAlias = TaskOutput

# This should be the same for all taskmodules
TaskEncodingType: TypeAlias = TaskEncoding[DocumentType, InputEncodingType, TargetEncodingType]
TaskModuleType: TypeAlias = TaskModule[
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
    ModelEncodingType,
    ModelOutputType,
    TaskOutputType,
]


@TaskModule.register()
class SimpleTransformerTextClassificationTaskModule(TaskModuleType):
    # If these attributes are set, the taskmodule is considered as prepared. They should be calculated
    # within _prepare() and are dumped automatically when saving the taskmodule with save_pretrained().
    PREPARED_ATTRIBUTES = ["label_to_id"]
    DOCUMENT_TYPE = TextDocumentWithLabel

    def __init__(
        self,
        tokenizer_name_or_path: str,
        padding: Union[bool, str, PaddingStrategy] = True,
        truncation: Union[bool, str, TruncationStrategy] = True,
        max_length: Optional[int] = None,
        pad_to_multiple_of: Optional[int] = None,
        label_to_id: Optional[Dict[str, int]] = None,
        **kwargs,
    ) -> None:
        # Important: Remaining keyword arguments need to be passed to super.
        super().__init__(**kwargs)
        # Save all passed arguments. They will be available via self._config().
        self.save_hyperparameters()

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

        # some tokenization and padding parameters
        self.truncation = truncation
        self.max_length = max_length
        self.padding = padding
        self.pad_to_multiple_of = pad_to_multiple_of

        # this will be prepared from the data or loaded from the config
        self.label_to_id = label_to_id

    def _prepare(self, documents: Sequence[DocumentType]) -> None:
        """
        Prepare the task module with training documents, e.g. collect all possible labels.
        This method needs to set all attributes listed in PREPARED_ATTRIBUTES.
        """

        # create the label-to-id mapping
        labels = set()
        for document in documents:
            # all annotations of a document are hold in list like containers,
            # so we have to take its first element
            label_annotation = document.label[0]
            labels.add(label_annotation.label)

        # create the mapping, but spare the first index for the "O" (outside) class
        self.label_to_id = {label: i + 1 for i, label in enumerate(sorted(labels))}
        self.label_to_id["O"] = 0

    def _post_prepare(self):
        """
        Any further preparation logic that requires the result of _prepare(). But its result is not serialized
        with the taskmodule.
        """

        self.id_to_label = {v: k for k, v in self.label_to_id.items()}

    def encode_input(
        self,
        document: DocumentType,
    ) -> TaskEncodingType:
        """
        Create one or multiple task encodings for the given document.
        """

        # tokenize the input text, this will be the input
        inputs = self.tokenizer(
            document.text,
            # we do not pad here, this will be done in collate()
            # when the actual batches are created
            padding=False,
            truncation=self.truncation,
            max_length=self.max_length,
        )

        return TaskEncoding(
            document=document,
            inputs=inputs,
        )

    def encode_target(
        self,
        task_encoding: TaskEncodingType,
    ) -> TargetEncodingType:
        """
        Create a target for a task encoding. This may use any annotations of the underlying document.
        """

        # as above, all annotations are hold in lists, so we have to take its first element
        label_annotation = task_encoding.document.label[0]
        # translate the textual label to the target id
        assert self.label_to_id is not None
        return self.label_to_id[label_annotation.label]

    def collate(self, task_encodings: Sequence[TaskEncodingType]) -> ModelEncodingType:
        """
        Convert a list of task encodings to a batch that will be passed to the model.
        """
        # get the inputs from the task encodings
        input_features = [task_encoding.inputs for task_encoding in task_encodings]

        # pad the inputs and return torch tensors
        inputs = self.tokenizer.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        if task_encodings[0].has_targets:
            # convert the targets (label ids) to a tensor
            targets = torch.tensor(
                [task_encoding.targets for task_encoding in task_encodings], dtype=torch.int64
            )
        else:
            # during inference, we do not have any targets
            targets = None

        return inputs, targets

    def unbatch_output(self, model_output: ModelOutputType) -> Sequence[TaskOutputType]:
        """
        Convert one model output batch to a sequence of taskmodule outputs.
        """

        # get the logits from the model output
        logits = model_output["logits"]

        # convert the logits to "probabilities"
        probabilities = logits.softmax(dim=-1).detach().cpu().numpy()

        # get the max class index per example
        max_label_ids = np.argmax(probabilities, axis=-1)

        outputs = []
        for idx, label_id in enumerate(max_label_ids):
            # translate the label id back to the label text
            label = self.id_to_label[label_id]
            # get the probability and convert from tensor value to python float
            prob = float(probabilities[idx, label_id])
            # we create TransformerTextClassificationTaskOutput primarily for typing purposes,
            # a simple dict would also work
            result: TaskOutput = {
                "label": label,
                "probability": prob,
            }
            outputs.append(result)

        return outputs

    def create_annotations_from_output(
        self,
        task_encodings: TaskEncodingType,
        task_outputs: TaskOutputType,
    ) -> Iterator[Tuple[str, Label]]:
        """
        Convert a task output to annotations. The method has to yield tuples (annotation_name, annotation).
        """

        # just yield a single annotation (other tasks may need multiple annotations per task output)
        yield "label", Label(label=task_outputs["label"], score=task_outputs["probability"])



"""
workflow:
    document
        -> (input_encoding, target_encoding) -> task_encoding
            -> model_encoding -> model_output
        -> task_output
    -> document
"""

import logging
import re
from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple, Type, Union

from transformers import AutoTokenizer
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import TruncationStrategy
from typing_extensions import TypeAlias

from pytorch_ie.annotations import BinaryRelation, LabeledSpan
from pytorch_ie.core import Annotation, TaskEncoding, TaskModule
from pytorch_ie.documents import TextDocument, TextDocumentWithLabeledSpansAndBinaryRelations
from pytorch_ie.models.transformer_seq2seq import ModelOutputType, ModelStepInputType

InputEncodingType: TypeAlias = Dict[str, Sequence[int]]
TargetEncodingType: TypeAlias = Dict[str, Sequence[int]]

TaskEncodingType: TypeAlias = TaskEncoding[TextDocument, InputEncodingType, TargetEncodingType]
TaskOutputType: TypeAlias = Sequence[Dict[str, Any]]

TaskModuleType: TypeAlias = TaskModule[
    TextDocument,
    InputEncodingType,
    TargetEncodingType,
    ModelStepInputType,
    ModelOutputType,
    TaskOutputType,
]

logger = logging.getLogger(__name__)


@TaskModule.register()
class TransformerSeq2SeqTaskModule(TaskModuleType):

    DOCUMENT_TYPE = TextDocumentWithLabeledSpansAndBinaryRelations

    def __init__(
        self,
        tokenizer_name_or_path: str,
        entity_annotation: str = "labeled_spans",
        relation_annotation: str = "binary_relations",
        padding: Union[bool, str, PaddingStrategy] = True,
        truncation: Union[bool, str, TruncationStrategy] = True,
        max_input_length: Optional[int] = None,
        max_target_length: Optional[int] = None,
        pad_to_multiple_of: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.relation_annotation = relation_annotation
        self.entity_annotation = entity_annotation
        self.padding = padding
        self.truncation = truncation
        self.max_input_length = max_input_length
        self.max_target_length = max_target_length
        self.pad_to_multiple_of = pad_to_multiple_of

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

    @property
    def document_type(self) -> Optional[Type[TextDocument]]:
        dt: Type[TextDocument] = self.DOCUMENT_TYPE
        if (
            self.entity_annotation == "labeled_spans"
            and self.relation_annotation == "binary_relations"
        ):
            return dt
        else:
            logger.warning(
                f"entity_annotation={self.entity_annotation} and relation_annotation={self.relation_annotation} are "
                f"not the default values ('labeled_spans' and 'binary_relations'), so the taskmodule "
                f"{type(self).__name__} can not request the usual document type ({dt.__name__}) for auto-conversion "
                f"because this has the bespoken default values as layer names instead of the provided ones."
            )
            return None

    def encode_text(self, text: str) -> InputEncodingType:
        return self.tokenizer(
            text,
            padding=False,
            truncation=self.truncation,
            max_length=self.max_input_length,
            is_split_into_words=False,
        )

    def encode_input(
        self,
        document: TextDocument,
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType],]]:
        return TaskEncoding(
            document=document,
            inputs=self.encode_text(document.text),
        )

    def document_to_target_string(self, document: TextDocument) -> str:
        relations: Sequence[BinaryRelation] = document[self.relation_annotation]

        head_to_tail_and_label: Dict[LabeledSpan, List[Tuple[LabeledSpan, str]]] = {}
        for relation in relations:
            if not isinstance(relation.head, LabeledSpan) or not isinstance(
                relation.tail, LabeledSpan
            ):
                raise TypeError(
                    f"the taskmodule expects the relation arguments to be of type LabeledSpan, "
                    f"but got {type(relation.head)} and {type(relation.tail)}"
                )
            if relation.head not in head_to_tail_and_label:
                head_to_tail_and_label[relation.head] = []

            head_to_tail_and_label[relation.head].append((relation.tail, relation.label))

        lin_triplets: List[str] = []
        for head in sorted(head_to_tail_and_label.keys(), key=lambda head: head.start):
            tail_and_label_for_head = head_to_tail_and_label[head]

            head_entity = document.text[head.start : head.end]

            lin_triplets.append("<triplet>")
            lin_triplets.append(head_entity)

            for tail, label in sorted(
                tail_and_label_for_head, key=lambda tail_and_label: tail_and_label[0].start
            ):
                tail_entity = document.text[tail.start : tail.end]

                lin_triplets.append("<subj>")
                lin_triplets.append(tail_entity)
                lin_triplets.append("<obj>")
                lin_triplets.append(label)

        return " ".join(lin_triplets)

    def encode_target(
        self,
        task_encoding: TaskEncodingType,
    ) -> TargetEncodingType:
        target_string = self.document_to_target_string(task_encoding.document)
        return {"labels": self.encode_text(target_string)["input_ids"]}

    def unbatch_output(self, model_output: ModelOutputType) -> Sequence[TaskOutputType]:
        unbatched_output = []
        for out in model_output:
            decoded_string = self.tokenizer.decode(
                out, skip_special_tokens=False, clean_up_tokenization_spaces=True
            )
            triplets = self._extract_triplets(decoded_string)
            unbatched_output.append(triplets)

        return unbatched_output

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncodingType,
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, Annotation]]:
        for relation_dct in task_output:
            head_entity = relation_dct["head"]
            tail_entity = relation_dct["tail"]
            label = relation_dct["type"]

            if label == "no_relation":
                continue

            # for now, just use the first head and tail match in the document
            text = task_encoding.document.text.lower()
            try:
                # this may fail if head_entity or tail_entity contains any special regex character (e.g. brackets)
                head_match = re.search(head_entity.lower(), text)
                tail_match = re.search(tail_entity.lower(), text)
            except Exception:
                logger.warning(
                    f"could not successfully search for the entities in the text, skip the triplet "
                    f'(head: "{head_entity}", tail: "{tail_entity}", label: "{label}")'
                )
                continue

            if head_match is None or tail_match is None:
                continue

            head = LabeledSpan(start=head_match.start(), end=head_match.end(), label="head")
            tail = LabeledSpan(start=tail_match.start(), end=tail_match.end(), label="tail")

            relation = BinaryRelation(head=head, tail=tail, label=label)

            yield from [
                (self.entity_annotation, head),
                (self.entity_annotation, tail),
                (self.relation_annotation, relation),
            ]

    def collate(self, task_encodings: Sequence[TaskEncodingType]) -> ModelStepInputType:
        input_features = [task_encoding.inputs for task_encoding in task_encodings]

        padded_encoding = self.tokenizer.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_input_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        if task_encodings[0].has_targets:
            # TODO: this is a bit of a hack -- fix
            labels = {
                "input_ids": [task_encoding.targets["labels"] for task_encoding in task_encodings]
            }

            padded_labels = self.tokenizer.pad(
                labels,
                padding=self.padding,
                max_length=self.max_target_length,
                pad_to_multiple_of=self.pad_to_multiple_of,
                return_tensors="pt",
            )

            # TODO: this is a bit of a hack -- fix
            padded_encoding["labels"] = padded_labels["input_ids"]

        return (padded_encoding,)

    # TODO: improve this method as soon as we have unittests for this taskmodule
    def _extract_triplets(self, text) -> TaskOutputType:
        triplets = []
        relation, subject, relation, object_ = "", "", "", ""
        text = text.strip()
        current = "x"
        for token in text.replace("<s>", "").replace("<pad>", "").replace("</s>", "").split():
            if token == "<triplet>":
                current = "t"
                if relation != "":
                    triplets.append(
                        {
                            "head": subject.strip(),
                            "type": relation.strip(),
                            "tail": object_.strip(),
                        }
                    )
                    relation = ""
                subject = ""
            elif token == "<subj>":
                current = "s"
                if relation != "":
                    triplets.append(
                        {
                            "head": subject.strip(),
                            "type": relation.strip(),
                            "tail": object_.strip(),
                        }
                    )
                object_ = ""
            elif token == "<obj>":
                current = "o"
                relation = ""
            else:
                if current == "t":
                    subject += " " + token
                elif current == "s":
                    object_ += " " + token
                elif current == "o":
                    relation += " " + token
        if subject != "" and relation != "" and object_ != "":
            triplets.append(
                {"head": subject.strip(), "type": relation.strip(), "tail": object_.strip()}
            )
        return triplets



import logging
from typing import Any, Dict, MutableMapping, Optional, Tuple

import torchmetrics
from torch import Tensor, nn
from torch.optim import AdamW
from transformers import AutoConfig, AutoModel, get_linear_schedule_with_warmup
from typing_extensions import TypeAlias

from pytorch_ie.core import PyTorchIEModel
from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses

ModelInputType: TypeAlias = MutableMapping[str, Any]
ModelOutputType: TypeAlias = Dict[str, Any]

ModelStepInputType = Tuple[
    ModelInputType,
    Optional[Tensor],
]

TRAINING = "train"
VALIDATION = "val"
TEST = "test"

logger = logging.getLogger(__name__)


@PyTorchIEModel.register()
class TransformerTextClassificationModel(
    PyTorchIEModel, RequiresModelNameOrPath, RequiresNumClasses
):
    def __init__(
        self,
        model_name_or_path: str,
        num_classes: int,
        tokenizer_vocab_size: Optional[int] = None,
        ignore_index: Optional[int] = None,
        learning_rate: float = 1e-5,
        task_learning_rate: float = 1e-4,
        warmup_proportion: float = 0.1,
        freeze_model: bool = False,
        multi_label: bool = False,
        t_total: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if t_total is not None:
            logger.warning(
                "t_total is deprecated, we use estimated_stepping_batches from the pytorch lightning trainer instead"
            )

        self.save_hyperparameters(ignore=["t_total"])

        self.learning_rate = learning_rate
        self.task_learning_rate = task_learning_rate
        self.warmup_proportion = warmup_proportion

        config = AutoConfig.from_pretrained(model_name_or_path)
        if self.is_from_pretrained:
            self.model = AutoModel.from_config(config=config)
        else:
            self.model = AutoModel.from_pretrained(model_name_or_path, config=config)

        if freeze_model:
            for param in self.model.parameters():
                param.requires_grad = False

        if tokenizer_vocab_size is not None:
            self.model.resize_token_embeddings(tokenizer_vocab_size)

        classifier_dropout = (
            config.classifier_dropout
            if config.classifier_dropout is not None
            else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)

        self.classifier = nn.Linear(config.hidden_size, num_classes)

        self.loss_fct = nn.BCEWithLogitsLoss() if multi_label else nn.CrossEntropyLoss()

        self.f1 = nn.ModuleDict(
            {
                f"stage_{stage}": torchmetrics.F1Score(
                    num_classes=num_classes,
                    ignore_index=ignore_index,
                    task="multilabel" if multi_label else "multiclass",
                )
                for stage in [TRAINING, VALIDATION, TEST]
            }
        )

    def forward(self, inputs: ModelInputType) -> ModelOutputType:
        output = self.model(**inputs)

        hidden_state = output.last_hidden_state

        cls_embeddings = hidden_state[:, 0, :]
        logits = self.classifier(cls_embeddings)

        return {"logits": logits}

    def step(self, stage: str, batch: ModelStepInputType):
        inputs, target = batch
        assert target is not None, "target has to be available for training"

        logits = self(inputs)["logits"]

        loss = self.loss_fct(logits, target)

        self.log(f"{stage}/loss", loss, on_step=(stage == TRAINING), on_epoch=True, prog_bar=True)

        f1 = self.f1[f"stage_{stage}"]
        f1(logits, target)
        self.log(f"{stage}/f1", f1, on_step=False, on_epoch=True, prog_bar=True)

        return loss

    def training_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TRAINING, batch=batch)

    def validation_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=VALIDATION, batch=batch)

    def test_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TEST, batch=batch)

    def configure_optimizers(self):
        optimizer = AdamW(self.parameters(), lr=self.learning_rate)
        if self.warmup_proportion > 0.0:
            stepping_batches = self.trainer.estimated_stepping_batches
            scheduler = get_linear_schedule_with_warmup(
                optimizer, int(stepping_batches * self.warmup_proportion), stepping_batches
            )
            return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        else:
            return optimizer



import logging
from typing import Any, Dict, Iterable, Optional, Sequence, Tuple

import torch
import torchmetrics
from torch import Tensor, nn
from torch.optim import AdamW
from transformers import AutoConfig, AutoModel, BatchEncoding, get_linear_schedule_with_warmup
from typing_extensions import TypeAlias

from pytorch_ie.core import PyTorchIEModel
from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses
from pytorch_ie.models.modules.mlp import MLP

ModelInputType: TypeAlias = BatchEncoding
ModelOutputType: TypeAlias = Dict[str, Any]

ModelStepInputType: TypeAlias = Tuple[
    ModelInputType,
    Optional[Sequence[Sequence[Tuple[int, int, int]]]],
]


TRAINING = "train"
VALIDATION = "val"
TEST = "test"


logger = logging.getLogger(__name__)


@PyTorchIEModel.register()
class TransformerSpanClassificationModel(
    PyTorchIEModel, RequiresModelNameOrPath, RequiresNumClasses
):
    def __init__(
        self,
        model_name_or_path: str,
        num_classes: int,
        learning_rate: float = 1e-5,
        task_learning_rate: float = 1e-4,
        warmup_proportion: float = 0.1,
        ignore_index: int = 0,
        max_span_length: int = 8,
        span_length_embedding_dim: int = 150,
        t_total: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if t_total is not None:
            logger.warning(
                "t_total is deprecated, we use estimated_stepping_batches from the pytorch lightning trainer instead"
            )

        self.save_hyperparameters(ignore=["t_total"])

        self.t_total = t_total
        self.learning_rate = learning_rate
        self.task_learning_rate = task_learning_rate
        self.warmup_proportion = warmup_proportion
        self.max_span_length = max_span_length

        config = AutoConfig.from_pretrained(model_name_or_path)
        if self.is_from_pretrained:
            self.model = AutoModel.from_config(config=config)
        else:
            self.model = AutoModel.from_pretrained(model_name_or_path, config=config)

        classifier_dropout = (
            config.classifier_dropout
            if config.classifier_dropout is not None
            else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        # TODO: properly intialize!
        # self.classifier = nn.Linear(config.hidden_size * 2 + span_length_embedding_dim, num_classes)
        self.classifier = MLP(
            input_dim=config.hidden_size * 2 + span_length_embedding_dim,
            output_dim=num_classes,
            hidden_dim=150,
            num_layers=2,
        )

        self.span_length_embedding = nn.Embedding(
            num_embeddings=max_span_length, embedding_dim=span_length_embedding_dim
        )

        self.loss_fct = nn.CrossEntropyLoss()

        self.f1 = nn.ModuleDict(
            {
                f"stage_{stage}": torchmetrics.F1Score(
                    num_classes=num_classes, ignore_index=ignore_index, task="multiclass"
                )
                for stage in [TRAINING, VALIDATION, TEST]
            }
        )

    def _start_end_and_span_length_span_index(
        self, batch_size: int, max_seq_length: int, seq_lengths: Optional[Iterable[int]] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        if seq_lengths is None:
            seq_lengths = batch_size * [max_seq_length]

        start_indices = []
        end_indices = []
        span_lengths = []
        span_batch_index = []
        offsets = []
        for batch_index, seq_length in enumerate(seq_lengths):
            offset = max_seq_length * batch_index

            for span_length in range(1, self.max_span_length + 1):
                for start_index in range(seq_length + 1 - span_length):
                    end_index = start_index + span_length - 1

                    span_batch_index.append(batch_index)
                    start_indices.append(start_index)
                    end_indices.append(end_index)
                    span_lengths.append(span_length - 1)
                    offsets.append(offset)

        return (
            torch.tensor(start_indices),
            torch.tensor(end_indices),
            torch.tensor(span_lengths),
            torch.tensor(span_batch_index),
            torch.tensor(offsets),
        )

    # TODO: this should live in the taskmodule
    def _expand_target_tuples(
        self,
        target_tuples: Sequence[Sequence[Tuple[int, int, int]]],
        batch_size: int,
        max_seq_length: int,
        seq_lengths: Optional[Iterable[int]] = None,
    ) -> torch.Tensor:
        if seq_lengths is None:
            seq_lengths = batch_size * [max_seq_length]

        target = []
        for batch_index, seq_length in enumerate(seq_lengths):
            label_lookup = {
                (start, end): label for start, end, label in target_tuples[batch_index]
            }
            for span_length in range(1, self.max_span_length + 1):
                for start_index in range(seq_length + 1 - span_length):
                    end_index = start_index + span_length - 1

                    label = label_lookup.get((start_index, end_index), 0)
                    target.append(label)

        return torch.tensor(target)

    def forward(self, inputs: ModelInputType) -> ModelOutputType:
        output = self.model(**inputs)

        batch_size, seq_length, hidden_dim = output.last_hidden_state.shape
        hidden_state = output.last_hidden_state.view(batch_size * seq_length, hidden_dim)

        seq_lengths = None
        if "attention_mask" in inputs:
            seq_lengths = torch.sum(inputs["attention_mask"], dim=-1).detach().cpu()

        (
            start_indices,
            end_indices,
            span_length,
            batch_indices,
            offsets,
        ) = self._start_end_and_span_length_span_index(
            batch_size=batch_size, max_seq_length=seq_length, seq_lengths=seq_lengths
        )

        start_embedding = hidden_state[offsets + start_indices, :]
        end_embedding = hidden_state[offsets + end_indices, :]
        span_length_embedding = self.span_length_embedding(span_length.to(hidden_state.device))

        combined_embedding = torch.cat(
            (start_embedding, end_embedding, span_length_embedding), dim=-1
        )

        logits = self.classifier(self.dropout(combined_embedding))

        return {
            "logits": logits,
            "batch_indices": batch_indices,
            "start_indices": start_indices,
            "end_indices": end_indices,
        }

    def step(self, stage: str, batch: ModelStepInputType, batch_idx):
        inputs, target_tuples = batch
        assert target_tuples is not None, f"target has to be available for {stage}"

        output = self(inputs)

        logits = output["logits"]

        batch_size, seq_length = inputs["input_ids"].shape
        seq_lengths = None
        if "attention_mask" in inputs:
            seq_lengths = torch.sum(inputs["attention_mask"], dim=-1)

        # TODO: Why is this not happening in TransformerSpanClassificationTaskModule.collate?
        target = self._expand_target_tuples(
            target_tuples=target_tuples,
            batch_size=batch_size,
            max_seq_length=seq_length,
            seq_lengths=seq_lengths,
        )
        target = target.to(logits.device)

        loss = self.loss_fct(logits, target)

        self.log(f"{stage}/loss", loss, on_step=stage == TRAINING, on_epoch=True, prog_bar=True)

        f1 = self.f1[f"stage_{stage}"]
        f1(logits, target)
        self.log(f"{stage}/f1", f1, on_step=False, on_epoch=True, prog_bar=True)

        return loss

    def training_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TRAINING, batch=batch, batch_idx=batch_idx)

    def validation_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=VALIDATION, batch=batch, batch_idx=batch_idx)

    def test_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TEST, batch=batch, batch_idx=batch_idx)

    def configure_optimizers(self):
        param_optimizer = list(self.named_parameters())
        # TODO: this needs fixing (does not work models other than BERT)
        optimizer_grouped_parameters = [
            {"params": [p for n, p in param_optimizer if "bert" in n]},
            {
                "params": [p for n, p in param_optimizer if "bert" not in n],
                "lr": self.task_learning_rate,
            },
        ]
        optimizer = AdamW(optimizer_grouped_parameters, lr=self.learning_rate)
        if self.warmup_proportion > 0.0:
            stepping_batches = self.trainer.estimated_stepping_batches
            scheduler = get_linear_schedule_with_warmup(
                optimizer, int(stepping_batches * self.warmup_proportion), stepping_batches
            )
            return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        else:
            return optimizer



import dataclasses
import logging
import typing
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import _asdict_inner  # type: ignore
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)

logger = logging.getLogger(__name__)


def _enumerate_dependencies(
    resolved: List[str],
    dependency_graph: Dict[str, List[str]],
    nodes: List[str],
    current_path: Optional[Set[str]] = None,
):
    current_path = current_path or set()
    for node in nodes:
        if node in current_path:
            raise ValueError(f"circular dependency detected at node: {node}")
        if node not in resolved:
            # terminal nodes
            if node not in dependency_graph:
                resolved.append(node)
            # nodes with dependencies
            else:
                # enumerate all dependencies first, then append itself
                _enumerate_dependencies(
                    resolved,
                    dependency_graph,
                    nodes=dependency_graph[node],
                    current_path=current_path | {node},
                )
                resolved.append(node)


def _is_optional_type(t: typing.Type) -> bool:
    type_origin = typing.get_origin(t)
    type_args = typing.get_args(t)
    return type_origin is typing.Union and len(type_args) == 2 and type(None) in type_args


def _is_optional_annotation_type(t: typing.Type) -> bool:
    is_optional = _is_optional_type(t)
    if not is_optional:
        return False
    return _is_annotation_type(typing.get_args(t)[0])


def _is_annotation_type(t: Any) -> bool:
    return type(t) == type and issubclass(t, Annotation)


def _contains_annotation_type(t: Any) -> bool:
    if _is_annotation_type(t):
        return True
    type_args = typing.get_args(t)
    return any(_contains_annotation_type(type_arg) for type_arg in type_args)


def _is_tuple_of_annotation_types(t: Any) -> bool:
    type_args = typing.get_args(t)
    if typing.get_origin(t) == tuple and _is_annotation_type(type_args[0]):
        if not (
            type_args[1] == Ellipsis
            or all(issubclass(type_arg, Annotation) for type_arg in type_args)
        ):
            raise TypeError(
                f"only tuples that do not mix Annotations with other types are supported"
            )
        return True
    else:
        return False


def _get_reference_fields_and_container_types(
    annotation_class: typing.Type["Annotation"],
) -> Dict[str, Any]:
    containers: Dict[str, Any] = {}
    for field in dataclasses.fields(annotation_class):
        if field.name == "_targets":
            continue
        if isinstance(field.type, type):
            field_type = field.type
        else:
            field_type = typing.get_type_hints(annotation_class)[field.name]
        if not _contains_annotation_type(field_type):
            continue
        if _is_optional_annotation_type(field_type):
            containers[field.name] = typing.Optional
            continue
        if _is_annotation_type(field_type):
            containers[field.name] = None
            continue
        if _is_tuple_of_annotation_types(field_type):
            containers[field.name] = tuple
            continue
        annot_name = annotation_class.__name__
        raise TypeError(
            f"The type '{field_type}' of the field '{field.name}' from Annotation subclass '{annot_name}' can not "
            f"be handled automatically. For automatic handling, type constructs that contain any Annotation subclasses "
            f"need to be either (1) pure subclasses of Annotation, (2) tuples of Annotation subclasses, or their "
            f"optional variants (examples: 1) Span, 2) Tuple[Span, ...], 3) Optional[Span]). Is the defined type "
            f"really the one you want to use? If so, consider to overwrite "
            f"{annot_name}.asdict() and {annot_name}.fromdict() by your own."
        )

    return containers


def _get_annotation_fields(fields: List[dataclasses.Field]) -> Set[dataclasses.Field]:
    # this was broken, so we raise an exception for now
    # return {f for f in fields if typing.get_origin(f.type) is AnnotationLayer}
    raise Exception(
        "_get_annotation_fields() is broken, please use Document.annotation_fields() instead"
    )


def annotation_field(
    target: Optional[str] = None,
    targets: Optional[List[str]] = None,
    named_targets: Optional[Dict[str, str]] = None,
):
    """
    We allow 3 variants to pass targets:
    1) as single value `target: str`: this works if only one target is required,
    2) as list of target names `targets: List[str]`: this works if multiple targets are required, but the
        respective Annotation class does _not_ define `TARGET_NAMES` (this disallows the usage of named_targets
        in the Annotation class), and
    3) as a mapping `named_targets: Dict[str, str]` from entries in Annotation.TARGET_NAMES to field names of the
        Document: This should be used if the respective Annotation class defines `TARGET_NAMES`, and thus,
        makes use of the `named_targets` property.
    """
    target_names = None
    new_targets = []
    if target is not None:
        if targets is not None or named_targets is not None:
            raise ValueError(f"only one of target, targets or named_targets can be set")
        new_targets = [target]
    if targets is not None:
        if target is not None or named_targets is not None:
            raise ValueError(f"only one of target, targets or named_targets can be set")
        new_targets = targets
    if named_targets is not None:
        if target is not None or targets is not None:
            raise ValueError(f"only one of target, targets or named_targets can be set")
        new_targets = list(named_targets.values())
        target_names = list(named_targets.keys())
    return dataclasses.field(
        metadata=dict(targets=new_targets, target_names=target_names), init=False, repr=False
    )


# for now, we only have annotation lists and texts
TARGET_TYPE = Union["AnnotationLayer", str]


@dataclasses.dataclass(eq=True, frozen=True)
class Annotation:
    _targets: Optional[Tuple[TARGET_TYPE, ...]] = dataclasses.field(
        default=None, init=False, repr=False, hash=False
    )
    TARGET_NAMES: ClassVar[Optional[Tuple[str, ...]]] = None

    def set_targets(self, value: Optional[Tuple[TARGET_TYPE, ...]]):
        if value is not None and self._targets is not None:
            raise ValueError(
                f"Annotation already has assigned targets. Clear the "
                f"annotation list container or remove the annotation with pop() "
                f"to assign it to a new annotation list with other targets."
            )
        object.__setattr__(self, "_targets", value)

    @property
    def non_comparison_fields_and_values(self) -> Tuple[Tuple[str, Any], ...]:
        return tuple(
            (f.name, getattr(self, f.name)) for f in dataclasses.fields(self) if not f.compare
        )

    @property
    def _id(self) -> int:
        # also calculate the hash over the non-comparison fields for id creation because otherwise
        # these fields would not be considered (per default, hash=false for non-comparison fields)
        # and (de-)serialization would de-duplicate annotations that vary just in these fields
        return hash((self, self.non_comparison_fields_and_values))

    @property
    def target(self) -> Optional[TARGET_TYPE]:
        if self._targets is None:
            return None
        if len(self._targets) == 0:
            raise ValueError(f"annotation has no target")
        if len(self._targets) > 1:
            raise ValueError(
                f"annotation has multiple targets, target is not defined in this case"
            )
        return self._targets[0]

    @property
    def targets(self) -> Optional[Tuple[TARGET_TYPE, ...]]:
        return self._targets

    @property
    def named_targets(self) -> Dict[str, TARGET_TYPE]:
        if self._targets is None:
            raise ValueError(
                f"targets is not set (this annotation may be not yet attached to a document)"
            )
        if self.TARGET_NAMES is None:
            raise TypeError(f"no TARGET_NAMES defined")
        return {name: self._targets[i] for i, name in enumerate(self.TARGET_NAMES)}

    def _asdict(
        self,
        exclude_fields: Optional[List[str]] = None,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        result: List[Tuple[str, Any]] = []
        _exclude_fields = set(exclude_fields) if exclude_fields is not None else set()
        _exclude_fields.add("_targets")
        if overrides is not None:
            _exclude_fields.update(overrides)
            result.extend(overrides.items())
        for f in dataclasses.fields(self):
            if f.name in _exclude_fields:
                continue
            field_value = getattr(self, f.name)
            value = _asdict_inner(field_value, dict)
            result.append((f.name, value))
        dct = dict(result)
        dct["_id"] = self._id
        return dct

    def asdict(self) -> Dict[str, Any]:
        overrides = {}
        reference_fields_with_container_type = _get_reference_fields_and_container_types(
            type(self)
        )
        for field_name, container_type in reference_fields_with_container_type.items():
            if container_type is None:
                overrides[field_name] = getattr(self, field_name)._id
            elif container_type == typing.Optional:
                field_value = getattr(self, field_name)
                overrides[field_name] = None if field_value is None else field_value._id
            elif container_type == tuple:
                # save as list to be json compatible
                overrides[field_name] = [anno._id for anno in getattr(self, field_name)]
            else:
                raise Exception(f"unknown annotation container type: {container_type}")

        dct = self._asdict(overrides=overrides)
        return dct

    @classmethod
    def fromdict(
        cls,
        dct: Dict[str, Any],
        annotation_store: Optional[Dict[int, "Annotation"]] = None,
    ):
        tmp_dct = dict(dct)
        reference_fields_with_container_type = _get_reference_fields_and_container_types(cls)
        for field_name, container_type in reference_fields_with_container_type.items():
            if container_type is None:
                tmp_dct[field_name] = resolve_annotation(
                    tmp_dct[field_name], store=annotation_store
                )
            elif container_type == typing.Optional:
                if tmp_dct[field_name] is None:
                    tmp_dct[field_name] = None
                else:
                    tmp_dct[field_name] = resolve_annotation(
                        tmp_dct[field_name], store=annotation_store
                    )
            elif container_type == tuple:
                tmp_dct[field_name] = tuple(
                    resolve_annotation(anno_dct, store=annotation_store)
                    for anno_dct in tmp_dct[field_name]
                )
            else:
                raise Exception(f"unknown annotation container type: {container_type}")

        tmp_dct.pop("_id", None)
        return cls(**tmp_dct)

    @property
    def is_attached(self) -> bool:
        return self._targets is not None

    def copy(self, **overrides) -> "Annotation":
        """
        Create a detached copy of the annotation with the same values as the original.

        :param overrides: keyword arguments to override the values of the original annotation
        :return: a detached copy of the annotation
        """
        kwargs = {}
        for f in dataclasses.fields(self):
            if f.name == "_targets":
                continue
            kwargs[f.name] = getattr(self, f.name)
        kwargs.update(overrides)
        return type(self)(**kwargs)

    def copy_with_store(
        self, override_annotation_store: Dict[int, "Annotation"], invalid_annotation_ids: Set[int]
    ) -> Optional["Annotation"]:
        """
        Create a detached copy of the annotation, but replace references to other annotations
        with entries from an override annotation store.
        Note: For now, fields are only allowed to directly reference an annotation or a tuple of annotations,
         but not any nested data structure that contains annotations.

        Args:
            :param override_annotation_store: the annotation store to use, a mapping from original *annotation ids*
                to the new *annotations*
            :param invalid_annotation_ids: a set of annotation ids that are not valid anymore and so any Annotation that
                references one of these ids will be discarded
            :return: a detached copy of the annotation or None if the annotation contains references to invalid
                annotations
        """
        overrides: Dict[str, Any] = {}
        for f in dataclasses.fields(self):
            if f.name == "_targets":
                continue
            field_value = getattr(self, f.name)
            if isinstance(field_value, Annotation):
                if field_value._id in invalid_annotation_ids:
                    return None
                overrides[f.name] = override_annotation_store.get(field_value._id, field_value)
            elif isinstance(field_value, tuple):
                if any(
                    maybe_anno._id in invalid_annotation_ids
                    for maybe_anno in field_value
                    if isinstance(maybe_anno, Annotation)
                ):
                    return None
                overrides[f.name] = tuple(
                    override_annotation_store.get(maybe_anno._id, maybe_anno)
                    if isinstance(maybe_anno, Annotation)
                    else maybe_anno
                    for maybe_anno in field_value
                )
            elif isinstance(field_value, (int, float, str, bool, type(None))):
                continue
            else:
                raise Exception(f"unknown annotation field type: {type(field_value)}")

        return self.copy(**overrides)


T = TypeVar("T", covariant=False, bound="Annotation")


class BaseAnnotationList(Sequence[T]):
    def __init__(self, document: "Document", targets: List[str]):
        self._document = document
        self._targets = targets
        self._annotations: List[T] = []

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseAnnotationList):
            return NotImplemented

        return self._targets == other._targets and self._annotations == other._annotations

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, s: slice) -> List[T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[T, List[T]]:
        return self._annotations[index]

    def __len__(self) -> int:
        return len(self._annotations)

    def append(self, annotation: T) -> None:
        targets = tuple(getattr(self._document, target_name) for target_name in self._targets)
        annotation.set_targets(targets)
        self._annotations.append(annotation)

    def extend(self, annotations: Iterable[T]) -> None:
        for annotation in annotations:
            self.append(annotation)

    def __repr__(self) -> str:
        return f"BaseAnnotationList({str(self._annotations)})"

    def clear(self) -> List[T]:
        """
        Detach all annotations from the layer and return them.
        """
        result = list(self._annotations)
        for annotation in self._annotations:
            annotation.set_targets(None)
        self._annotations = []
        return result

    def pop(self, index: int = -1) -> T:
        ann = self._annotations.pop(index)
        ann.set_targets(None)
        return ann

    @property
    def targets(self) -> dict[str, Any]:
        return {
            target_field_name: getattr(self._document, target_field_name)
            for target_field_name in self._targets
        }

    @property
    def target_names(self) -> List[str]:
        return self._targets

    @property
    def target_name(self) -> str:
        if len(self._targets) != 1:
            raise ValueError(
                f"The annotation layer has more or less than one target, can not return a single target name: "
                f"{self._targets}"
            )
        return self._targets[0]

    @property
    def target(self) -> Any:
        return self.targets[self.target_name]

    @property
    def target_layers(self) -> dict[str, "AnnotationLayer"]:
        return {
            target_name: target
            for target_name, target in self.targets.items()
            if isinstance(target, AnnotationLayer)
        }

    @property
    def target_layer(self) -> "AnnotationLayer":
        tgt_layers = self.target_layers
        if len(tgt_layers) != 1:
            raise ValueError(
                f"The annotation layer has more or less than one target layer: {list(tgt_layers.keys())}"
            )
        return list(tgt_layers.values())[0]


class AnnotationLayer(BaseAnnotationList[T]):
    def __init__(self, document: "Document", targets: List["str"]):
        super().__init__(document=document, targets=targets)
        self._predictions: BaseAnnotationList[T] = BaseAnnotationList(document, targets=targets)

    @property
    def predictions(self) -> BaseAnnotationList[T]:
        return self._predictions

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AnnotationLayer):
            return NotImplemented

        return super().__eq__(other) and self.predictions == other.predictions

    def __repr__(self) -> str:
        return f"AnnotationLayer({str(self._annotations)})"


D = TypeVar("D", bound="Document")


@dataclasses.dataclass
class Document(Mapping[str, Any]):
    # points from annotation field names to lists of target field names
    _annotation_graph: Dict[str, List[str]] = dataclasses.field(
        default_factory=dict, init=False, repr=False
    )
    _annotation_fields: Set[str] = dataclasses.field(default_factory=set, init=False, repr=False)

    @classmethod
    def fields(cls):
        return [
            f
            for f in dataclasses.fields(cls)
            if f.name not in {"_annotation_graph", "_annotation_fields"}
        ]

    @classmethod
    def field_types(cls) -> Dict[str, typing.Type]:
        result = {}
        for f in cls.fields():
            # If we got just the string representation of the type, we resolve the whole class.
            # But this may be slow, so we only do it if necessary.
            if not isinstance(f.type, type):
                return typing.get_type_hints(cls)
            result[f.name] = f.type
        return result

    @classmethod
    def annotation_fields(cls) -> Set[dataclasses.Field]:
        ann_field_types = cls.field_types()
        return {
            f
            for f in cls.fields()
            if typing.get_origin(ann_field_types[f.name]) is AnnotationLayer
        }

    def __getitem__(self, key: str) -> AnnotationLayer:
        if key not in self._annotation_fields:
            raise KeyError(f"Document has no attribute '{key}'.")
        return getattr(self, key)

    def __iter__(self):
        return iter(self._annotation_fields)

    def __len__(self):
        return len(self._annotation_fields)

    def __post_init__(self):
        targeted = set()
        field_names = {field.name for field in dataclasses.fields(self)}
        field_types = self.field_types()
        for field in self.annotation_fields():

            self._annotation_fields.add(field.name)

            targets = field.metadata.get("targets")
            for target in targets:
                targeted.add(target)
                if field.name not in self._annotation_graph:
                    self._annotation_graph[field.name] = []
                self._annotation_graph[field.name].append(target)
                if target not in field_names:
                    raise TypeError(
                        f'annotation target "{target}" is not in field names of the document: {field_names}'
                    )

            # check annotation target names and use them together with target names from the AnnotationLayer
            # to reorder targets, if available
            target_names = field.metadata.get("target_names")
            field_type = field_types[field.name]
            annotation_type = typing.get_args(field_type)[0]
            annotation_target_names = annotation_type.TARGET_NAMES
            if annotation_target_names is not None:
                if target_names is not None:
                    if set(target_names) != set(annotation_target_names):
                        raise TypeError(
                            f"keys of targets {sorted(target_names)} do not match "
                            f"{annotation_type.__name__}.TARGET_NAMES {sorted(annotation_target_names)}"
                        )
                    # reorder targets according to annotation_target_names
                    target_name_mapping = dict(zip(target_names, targets))
                    target_position_mapping = {
                        i: target_name_mapping[name]
                        for i, name in enumerate(annotation_target_names)
                    }
                    targets = [target_position_mapping[i] for i in range(len(targets))]
                else:
                    if len(annotation_target_names) != len(targets):
                        raise TypeError(
                            f"number of targets {sorted(targets)} does not match number of entries in "
                            f"{annotation_type.__name__}.TARGET_NAMES: {sorted(annotation_target_names)}"
                        )
                    # disallow multiple targets when target names are specified in the definition of the Annotation
                    if len(annotation_target_names) > 1:
                        raise TypeError(
                            f"A target name mapping is required for AnnotationLayers containing Annotations with "
                            f'TARGET_NAMES, but AnnotationLayer "{field.name}" has no target_names. You should '
                            f"pass the named_targets dict containing the following keys (see Annotation "
                            f'"{annotation_type.__name__}") to annotation_field: {annotation_target_names}'
                        )

            field_value = field_type(document=self, targets=targets)
            setattr(self, field.name, field_value)

        if "_artificial_root" in self._annotation_graph:
            raise ValueError(
                'Failed to add the "_artificial_root" node to the annotation graph because it already exists. Note '
                "that AnnotationLayer entries with that name are not allowed."
            )
        self._annotation_graph["_artificial_root"] = list(self._annotation_fields - targeted)

    def asdict(self):
        dct = {}
        for field in self.fields():
            value = getattr(self, field.name)

            if isinstance(value, AnnotationLayer):
                dct[field.name] = {
                    "annotations": [v.asdict() for v in value],
                    "predictions": [v.asdict() for v in value.predictions],
                }
            elif isinstance(value, dict):
                dct[field.name] = value or None
            else:
                dct[field.name] = value

        return dct

    @classmethod
    def fromdict(cls, dct):
        fields = dataclasses.fields(cls)
        annotation_fields = cls.annotation_fields()
        field_types = cls.field_types()

        cls_kwargs = {}
        for field in fields:
            if field not in annotation_fields:
                value = dct.get(field.name)

                if value is not None:
                    cls_kwargs[field.name] = value

        doc = cls(**cls_kwargs)

        name_to_field = {f.name: f for f in annotation_fields}

        dependency_ordered_fields: List[str] = []
        _enumerate_dependencies(
            dependency_ordered_fields,
            dependency_graph=doc._annotation_graph,
            nodes=doc._annotation_graph["_artificial_root"],
        )

        annotations = {}
        predictions = {}
        annotations_per_field = defaultdict(list)
        predictions_per_field = defaultdict(list)
        for field_name in dependency_ordered_fields:
            # terminal nodes do not have to be an annotation field (e.g. the text field)
            if field_name not in name_to_field:
                continue

            field = name_to_field[field_name]

            value = dct.get(field.name)

            if value is None or not value:
                continue

            field_type = field_types[field_name]
            # TODO: handle single annotations, e.g. a document-level label
            if typing.get_origin(field_type) is AnnotationLayer:
                annotation_class = typing.get_args(field_type)[0]
                # build annotations
                for annotation_data in value["annotations"]:
                    annotation_dict = dict(annotation_data)
                    annotation_id = annotation_dict.pop("_id")
                    # annotations can only reference annotations
                    annotation = annotation_class.fromdict(annotation_dict, annotations)
                    annotations[annotation_id] = annotation
                    annotations_per_field[field.name].append(annotation)
                # build predictions
                for annotation_data in value["predictions"]:
                    annotation_dict = dict(annotation_data)
                    annotation_id = annotation_dict.pop("_id")
                    # predictions can reference annotations and predictions
                    annotation = annotation_class.fromdict(
                        annotation_dict, {**annotations, **predictions}
                    )
                    predictions[annotation_id] = annotation
                    predictions_per_field[field.name].append(annotation)
            else:
                raise Exception("Error")

        for field_name, field_annotations in annotations_per_field.items():
            getattr(doc, field_name).extend(field_annotations)

        for field_name, field_annotations in predictions_per_field.items():
            getattr(doc, field_name).predictions.extend(field_annotations)

        return doc

    def as_type(
        self,
        new_type: typing.Type[D],
        field_mapping: Optional[Dict[str, str]] = None,
        keep_remaining: bool = True,
    ) -> D:
        field_mapping = field_mapping or {}
        new_doc = new_type.fromdict(
            {
                field_mapping.get(k, k): v
                for k, v in self.asdict().items()
                if keep_remaining or k in field_mapping
            }
        )
        return new_doc

    def copy(self, with_annotations: bool = True) -> "Document":
        doc_dict = self.asdict()
        if not with_annotations:
            for field in self.annotation_fields():
                doc_dict.pop(field.name)
        return type(self).fromdict(doc_dict)

    def add_all_annotations_from_other(
        self,
        other: "Document",
        removed_annotations: Optional[Dict[str, Set[int]]] = None,
        override_annotations: Optional[Dict[str, Dict[int, Annotation]]] = None,
        process_predictions: bool = True,
        strict: bool = True,
        verbose: bool = True,
    ) -> Dict[str, List[Annotation]]:
        """Adds all annotations from another document to this document. It allows to blacklist annotations
        and also to override annotations. It returns the original annotations for which a new annotation was
        added to the current document.

        The method is useful if e.g. a text-based document is converted to a token-based document and the
        annotations should be added to the token-based document.

        Args:
            other: The document to add annotations from.
            process_predictions: Whether to process predictions as well (default: True). If set to False,
                the predictions in the other document will be ignored.
            override_annotations: A mapping from annotation field names to a mapping from
                annotation IDs to annotations. The effects are two-fold. First, adding any annotation
                field name as key has the effect that the field is expected to be already handled and
                no annotations will be added from the other document. Second, if a certain mapping
                old_annotation._id -> new_annotation is present in the mapping for a certain field,
                the new_annotation will be used anywhere where the old_annotation would have been
                referenced. This propagates along the annotation graph and can be useful if some
                annotations are modified, but all dependent annotations should be kept intact e.g.
                when converting a text-based document to token-based (or the other way around). See below
                for an example.
            removed_annotations: A mapping from annotation field names to a set of annotation ids that
                are removed from the document. This is useful if e.g. the annotation base (the text)
                is split up and the annotations need to be split up as well or if conversion to a token base
                is performed, but some spans could not be converted, because of aligning issues, and, thus,
                should be removed. Similar as for entries in override_annotations, fields that are mentioned
                in removed_annotations are expected to be already handled manually and will not be added from
                the other document. Also, the removal of annotations propagates along the annotation graph.
            strict: Whether to raise an exception if the other document contains annotations that reference
                annotations that are not present in the current document (see parameter removed_annotations).
                If set to False, the annotations are ignored.
            verbose: Whether to print a warning if the other document contains annotations that reference
                annotations that are not present in the current document (see parameter removed_annotations).

        Returns:
            A mapping from annotation field names to the set of annotations from the original document for which
            a new annotation was added to the current document. This can be useful to check if all original
            annotations were added (possibly to multiple target documents).

        Example:
                ```
                @dataclasses.dataclass(frozen=True)
                class Attribute(Annotation):
                    ref: Annotation
                    value: str

                @dataclasses.dataclass
                class TextBasedDocumentWithEntitiesRelationsAndRelationAttributes(TextBasedDocument):
                    entities: AnnotationLayer[LabeledSpan] = annotation_field(target="text")
                    relations: AnnotationLayer[BinaryRelation] = annotation_field(target="entities")
                    relation_attributes: AnnotationLayer[Attribute] = annotation_field(target="relations")

                @dataclasses.dataclass
                class TokenBasedDocumentWithEntitiesRelationsAndRelationAttributes(TokenBasedDocument):
                    entities: AnnotationLayer[LabeledSpan] = annotation_field(target="tokens")
                    relations: AnnotationLayer[BinaryRelation] = annotation_field(target="entities")
                    relation_attributes: AnnotationLayer[Attribute] = annotation_field(target="relations")


                doc_text = TextBasedDocumentWithEntitiesRelationsAndRelationAttributes(text="Hello World!")
                e1_text = LabeledSpan(0, 5, "word1")
                e2_text = LabeledSpan(6, 11, "word2")
                doc_text.entities.extend([e1, e2])
                r1_text = BinaryRelation(e1_text, e2_text, "relation1")
                doc_text.relations.append(r1_text)
                doc_text.relation_attributes.append(Attribute(r1_text, "attribute1"))

                doc_tokens = TokenBasedDocumentWithEntitiesRelationsAndRelationAttributes(
                    tokens=("Hello", "World", "!")
                )
                e1_tokens = LabeledSpan(0, 1, "word1")
                e2_tokens = LabeledSpan(1, 2, "word2")
                doc_tokens.entities.extend([e1_tokens, e2_tokens])
                doc_tokens.add_all_annotations_from_other(
                    other=doc_text,
                    override_annotations={"entities": {e1_text._id: e1_tokens, e2_text._id: e2_tokens}},
                )
                # Note that the relation and attribute are still present, but now refer to the new entities
                # and new relation, respectively.
                ```
        """
        removed_annotations = defaultdict(set, removed_annotations or dict())
        added_annotations = defaultdict(list)

        annotation_store: Dict[str, Dict[int, Annotation]] = defaultdict(dict)
        named_annotation_fields = {field.name: field for field in self.annotation_fields()}
        if override_annotations is not None:
            for field_name, mapping in override_annotations.items():
                if field_name not in named_annotation_fields:
                    raise ValueError(
                        f'Field "{field_name}" is not an annotation field of {type(self).__name__}, but keys in '
                        f"override_annotation_mapping must be annotation field names."
                    )
                annotation_store[field_name].update(mapping)
        else:
            override_annotations = dict()

        dependency_ordered_fields: List[str] = []
        _enumerate_dependencies(
            dependency_ordered_fields,
            dependency_graph=self._annotation_graph,
            nodes=self._annotation_graph["_artificial_root"],
        )
        for field_name in dependency_ordered_fields:
            # we process only annotation fields that are not in the override_annotations and the removed_annotations
            # mapping because they are meant to be already manually handled
            if field_name in named_annotation_fields and field_name not in (
                set(override_annotations) | set(removed_annotations)
            ):
                current_targets_store = dict()
                current_invalid_annotation_ids = set()
                for target in self._annotation_graph.get(field_name, []):
                    current_targets_store.update(annotation_store[target])
                    current_invalid_annotation_ids.update(removed_annotations.get(target, set()))

                other_annotation_field = other[field_name]
                for ann in other_annotation_field:
                    new_ann = ann.copy_with_store(
                        override_annotation_store=current_targets_store,
                        invalid_annotation_ids=current_invalid_annotation_ids,
                    )
                    if new_ann is not None:
                        if ann._id != new_ann._id:
                            annotation_store[field_name][ann._id] = new_ann
                        self[field_name].append(new_ann)
                        added_annotations[field_name].append(ann)
                    else:
                        if strict:
                            raise ValueError(
                                f"Could not add annotation {ann} to {type(self).__name__} because it depends on "
                                f"annotations that are not present in the document."
                            )
                        if verbose:
                            logger.warning(
                                f"Could not add annotation {ann} to {type(self).__name__} because it depends on "
                                f"annotations that are not present in the document. The annotation is ignored."
                                f"(disable this warning with verbose=False)"
                            )
                        # The annotation was removed, so we need to make sure that it is not referenced by any other
                        removed_annotations[field_name].add(ann._id)
                if process_predictions:
                    for ann in other_annotation_field.predictions:
                        new_ann = ann.copy_with_store(
                            override_annotation_store=current_targets_store,
                            invalid_annotation_ids=current_invalid_annotation_ids,
                        )
                        if new_ann is not None:
                            if ann._id != new_ann._id:
                                annotation_store[field_name][ann._id] = new_ann
                            self[field_name].predictions.append(new_ann)
                            added_annotations[field_name].append(ann)
                        else:
                            if strict:
                                raise ValueError(
                                    f"Could not add annotation {ann} to {type(self).__name__} because it depends on "
                                    f"annotations that are not present in the document."
                                )
                            if verbose:
                                logger.warning(
                                    f"Could not add annotation {ann} to {type(self).__name__} because it depends on "
                                    f"annotations that are not present in the document. The annotation is ignored. "
                                    f"(disable this warning with verbose=False)"
                                )
                            # The annotation was removed, so we need to make sure that it is not referenced by any other
                            removed_annotations[field_name].add(ann._id)

        return dict(added_annotations)


def resolve_annotation(
    id_or_annotation: Union[int, Annotation],
    store: Optional[Dict[int, Annotation]],
) -> Annotation:
    if isinstance(id_or_annotation, Annotation):
        return id_or_annotation
    else:
        if store is None:
            raise ValueError("Unable to resolve the annotation id without annotation_store.")
        return store[id_or_annotation]



from .document import Annotation, AnnotationLayer, Document, annotation_field
from .metric import DocumentMetric
from .model import PyTorchIEModel
from .module_mixins import (
    EnterDatasetDictMixin,
    EnterDatasetMixin,
    ExitDatasetDictMixin,
    ExitDatasetMixin,
    PreparableMixin,
    WithDocumentTypeMixin,
)
from .statistic import DocumentStatistic
from .taskmodule import TaskEncoding, TaskModule

# backwards compatibility
AnnotationList = AnnotationLayer
RequiresDocumentTypeMixin = WithDocumentTypeMixin



from typing import Iterator, Optional, Sequence, Tuple


def enumerate_windows(
    sequence: Sequence, max_size, overlap: int = 0
) -> Iterator[Tuple[Tuple[int, int], Tuple[int, int]]]:
    """
    Enumerate all windows as slices over a sequence, optionally with an overlap. Overlap is interpreted as number of
    tokens taken into account that are already part in another window. We also return label_offset_slice that defines
    the slice (with respect to the token_slice!) of tokens that are not available in another slice.
    """
    window_without_overlap = max_size - 2 * overlap
    for label_start_idx in range(overlap, len(sequence), window_without_overlap):
        token_start_idx = label_start_idx - overlap
        label_end_idx = min(label_start_idx + window_without_overlap, len(sequence))
        token_end_idx = min(label_end_idx + overlap, len(sequence))
        label_start_offset = label_start_idx - token_start_idx
        label_end_offset = label_end_idx - token_start_idx
        token_slice = (token_start_idx, token_end_idx)
        # also allow using previous/remaining entries as labels if we are at the beginning/end
        # to cover all entries exactly once in a label slice
        if token_start_idx == 0:
            label_start_offset = 0
        if token_end_idx == len(sequence):
            label_end_offset = token_end_idx - token_start_idx
        label_offset_slice = (label_start_offset, label_end_offset)
        yield token_slice, label_offset_slice


def get_window_around_slice(
    slice: Tuple[int, int], max_window_size: int, available_input_length: int
) -> Optional[Tuple[int, int]]:
    """
    Given a `max_window` size, `available_token_length` and a `slice` (pair of start and end indices) that
    is required to be in the resulting window, create a new slice of size `max_window_size` (or less, if not possible)
    around the required slice. Per default, the resulting slice will be centered around the required slice.
    However, if the required slice is at the beginning or end of the available tokens, the resulting window is
    shifted to contain as many tokens as possible.
    Iff the required `slice` already exceeds the `max_window_size`, return `None`.
    """

    # current pair may not fit into the window
    if slice[1] - slice[0] > max_window_size:
        return None

    # set the final window size (regarding input tokens)
    window_size = min(available_input_length, max_window_size)

    rel_center = sum(slice) / 2.0
    window_start = int(rel_center - window_size / 2.0)
    window_end = window_start + window_size

    # If window goes over one end, shift it use as much content as possible.
    # First shift window to left and then to right to ensure that window_start is never
    # negative (if window_end is outside, this will not be a problem)
    if window_end >= available_input_length:
        delta = available_input_length - window_end
        window_start += delta
        window_end += delta
    if window_start < 0:
        delta = -window_start
        window_start += delta
        window_end += delta
    assert (
        0 <= window_start < available_input_length
    ), f"window_start={window_start} not available in sequence"

    return window_start, window_end
